from slp.modules.attention import Attention, MultiheadAttention, attention_scores
from slp.modules.classifier import Classifier
from slp.modules.embed import Embed, PositionalEncoding
from slp.modules.feedforward import FF, MultilayerFF, PositionwiseFF
from slp.modules.norm import LayerNorm
from slp.modules.regularization import GaussianNoise
from slp.modules.rnn import RNN, WordRNN
from slp.modules.transformer import Decoder as TransformerDecoder
from slp.modules.transformer import Encoder as TransformerEncoder
from slp.modules.transformer import Transformer



import argparse
from typing import Any, Callable, List, Optional, Union

import numpy as np
import pytorch_lightning as pl
import torch
from loguru import logger
from sklearn.model_selection import train_test_split
from slp.data.corpus import HfCorpus, TokenizedCorpus, WordCorpus
from slp.data.datasets import CorpusDataset, CorpusLMDataset
from slp.data.transforms import ToTensor
from slp.util.types import dir_path
from torch.utils.data import BatchSampler, DataLoader, Dataset, Sampler, random_split
from transformers import ALL_PRETRAINED_CONFIG_ARCHIVE_MAP

DatasetType = Union[Dataset, List[Any]]


def split_data(dataset, test_size, seed):
    """Train-test split of dataset.

    Dataset can be either a torch.utils.data.Dataset or a list

    Args:
        dataset (Union[Dataset, List]): Input dataset
        test_size (float): Size of the test set. Defaults to 0.2.
        seed (int): Optional seed for deterministic run. Defaults to None.

    Returns:
        Tuple[Union[Dataset, List], Union[Dataset, List]: (train set, test set)
    """
    train, test = None, None

    if isinstance(dataset, torch.utils.data.Dataset):
        test_len = int(test_size * len(dataset))
        train_len = len(dataset) - test_len

        seed_generator = None
        if seed is not None:
            seed_generator = torch.Generator().manual_seed(seed)

        train, test = random_split(
            dataset, [train_len, test_len], generator=seed_generator
        )

    else:

        train, test = train_test_split(dataset, test_size=test_size, random_state=seed)

    return train, test


class PLDataModuleFromDatasets(pl.LightningDataModule):
    def __init__(
        self,
        train: Dataset,
        val: Dataset = None,
        test: Dataset = None,
        val_percent: float = 0.2,
        test_percent: float = 0.2,
        batch_size: int = 1,
        batch_size_eval: Optional[int] = None,
        seed: Optional[int] = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        drop_last: bool = False,
        sampler_train: Sampler = None,
        sampler_val: Sampler = None,
        sampler_test: Sampler = None,
        batch_sampler_train: BatchSampler = None,
        batch_sampler_val: BatchSampler = None,
        batch_sampler_test: BatchSampler = None,
        shuffle_eval: bool = False,
        collate_fn: Optional[Callable[..., Any]] = None,
    ):
        """LightningDataModule wrapper for generic torch.utils.data.Dataset

        If val or test Datasets are not provided, this class will split
        val_pecent and test_percent of the train set respectively to create them

        Args:
            train (Dataset): Train set
            val (Dataset): Validation set. Defaults to None.
            test (Dataset): Test set. Defaults to None.
            val_percent (float): Percent of train to be used for validation if no validation set is given. Defaults to 0.2.
            test_percent (float): Percent of train to be used for test set if no test set is given. Defaults to 0.2.
            batch_size (int): Training batch size. Defaults to 1.
            batch_size_eval (Optional[int]): Validation and test batch size. Defaults to None.
            seed (Optional[int]): Seed for deterministic run. Defaults to None.
            num_workers (int): Number of workers in the DataLoader. Defaults to 1.
            pin_memory (bool): Pin tensors to GPU memory. Defaults to True.
            drop_last (bool): Drop last incomplete batch. Defaults to False.
            sampler_train (Sampler): Sampler for train loader. Defaults to None.
            sampler_val (Sampler): Sampler for validation loader. Defaults to None.
            sampler_test (Sampler): Sampler for test loader. Defaults to None.
            batch_sampler_train (BatchSampler): Batch sampler for train loader. Defaults to None.
            batch_sampler_val (BatchSampler): Batch sampler for validation loader. Defaults to None.
            batch_sampler_test (BatchSampler): Batch sampler for test loader. Defaults to None.
            shuffle_eval (bool): Shuffle validation and test dataloaders. Defaults to False.
            collate_fn (Callable[..., Any]): Collator function. Defaults to None.

        Raises:
            ValueError: If both mutually exclusive sampler_train and batch_sampler_train are provided
            ValueError: If both mutually exclusive sampler_val and batch_sampler_val are provided
            ValueError: If both mutually exclusive sampler_test and batch_sampler_test are provided
        """
        super(PLDataModuleFromDatasets, self).__init__()

        if batch_sampler_train is not None and sampler_train is not None:
            raise ValueError(
                "You provided both a sampler and a batch sampler for the train set. These are mutually exclusive"
            )

        if batch_sampler_val is not None and sampler_val is not None:
            raise ValueError(
                "You provided both a sampler and a batch sampler for the validation set. These are mutually exclusive"
            )
        if batch_sampler_test is not None and sampler_test is not None:
            raise ValueError(
                "You provided both a sampler and a batch sampler for the test set. These are mutually exclusive"
            )
        self.sampler_train = sampler_train
        self.sampler_val = sampler_val
        self.sampler_test = sampler_test
        self.batch_sampler_train = batch_sampler_train
        self.batch_sampler_val = batch_sampler_val
        self.batch_sampler_test = batch_sampler_test
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.drop_last = drop_last

        self.shuffle_eval = shuffle_eval
        self.collate_fn = collate_fn

        self.batch_size = batch_size

        if batch_size_eval is None:
            batch_size_eval = self.batch_size

        self.batch_size_eval = batch_size_eval

        self.train = train
        self.val = val
        self.test = test

        if self.val is not None:
            logger.info(f"Using provided validation set")

        if self.test is not None:
            logger.info(f"Using provided test set")

        if self.test is None and val is None:
            assert (
                val_percent is not None and val_percent > 0
            ), "You should either provide a validation set or a val set percentage"

            assert (
                test_percent is not None and test_percent > 0
            ), "You should either provide a test set or a test set percentage"

            testval_percent = test_percent + val_percent

            logger.info(
                f"No test or validation set provided. Creating random splits using {testval_percent * 100}% of training set with seed={seed}"
            )

            self.train, testval = split_data(
                self.train, test_size=testval_percent, seed=seed
            )

            test_percent = test_percent / testval_percent
            self.val, self.test = split_data(testval, test_size=test_percent, seed=seed)

        if self.val is None:
            assert (
                val_percent is not None and val_percent > 0
            ), "You should either provide a validation set or a val set percentage"

            logger.info(
                f"No validation set provided. Creating random split using {val_percent * 100}% of training set with seed={seed}"
            )
            self.train, self.val = split_data(
                self.train, test_size=val_percent, seed=seed
            )

        if self.test is None:
            assert (
                test_percent is not None and test_percent > 0
            ), "You should either provide a test set or a test set percentage"

            logger.info(
                f"No test set provided. Creating random split using {test_percent * 100}% of training set with seed={seed}"
            )
            self.train, self.test = split_data(
                self.train, test_size=test_percent, seed=seed
            )

        logger.info(f"Using {len(self.train)} samples for training")  # type: ignore
        logger.info(f"Using {len(self.val)} samples for validation")  # type: ignore
        logger.info(f"Using {len(self.test)} samples for testing")  # type: ignore

    def train_dataloader(self) -> DataLoader:
        """Configure train DataLoader

        Returns:
            DataLoader: Pytorch DataLoader for train set
        """
        return DataLoader(
            self.train,
            batch_size=self.batch_size if self.batch_sampler_train is None else 1,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last and (self.batch_sampler_train is None),
            sampler=self.sampler_train,
            batch_sampler=self.batch_sampler_train,
            shuffle=(self.batch_sampler_train is None) and (self.sampler_train is None),
            collate_fn=self.collate_fn,
        )

    def val_dataloader(self):
        """Configure validation DataLoader

        Returns:
            DataLoader: Pytorch DataLoader for validation set
        """
        val = DataLoader(
            self.val,
            batch_size=self.batch_size_eval if self.batch_sampler_val is None else 1,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last and (self.batch_sampler_val is None),
            sampler=self.sampler_val,
            batch_sampler=self.batch_sampler_val,
            shuffle=(
                self.shuffle_eval
                and (self.batch_sampler_val is None)
                and (self.sampler_val is None)
            ),
            collate_fn=self.collate_fn,
        )

        return val

    def test_dataloader(self):
        """Configure test DataLoader

        Returns:
            DataLoader: Pytorch DataLoader for test set
        """
        return DataLoader(
            self.test,
            batch_size=self.batch_size_eval if self.batch_sampler_test is None else 1,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last and (self.batch_sampler_test is None),
            sampler=self.sampler_test,
            batch_sampler=self.batch_sampler_test,
            shuffle=(
                self.shuffle_eval
                and (self.batch_sampler_test is None)
                and (self.sampler_test is None)
            ),
            collate_fn=self.collate_fn,
        )

    @classmethod
    def add_argparse_args(
        cls, parent_parser: argparse.ArgumentParser
    ) -> argparse.ArgumentParser:
        """Augment input parser with arguments for data loading

        Args:
            parent_parser (argparse.ArgumentParser): Parser created by the user

        Returns:
            argparse.ArgumentParser: Augmented parser
        """
        parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument(
            "--val-percent",
            dest="data.val_percent",
            type=float,
            help="Percent of validation data to be randomly split from the training set, if no validation set is provided",
        )

        parser.add_argument(
            "--test-percent",
            dest="data.test_percent",
            type=float,
            help="Percent of test data to be randomly split from the training set, if no test set is provided",
        )

        parser.add_argument(
            "--bsz", dest="data.batch_size", type=int, help="Training batch size"
        )

        parser.add_argument(
            "--bsz-eval",
            dest="data.batch_size_eval",
            type=int,
            help="Evaluation batch size",
        )

        parser.add_argument(
            "--num-workers",
            dest="data.num_workers",
            type=int,
            default=1,
            help="Number of workers to be used in the DataLoader",
        )

        parser.add_argument(
            "--pin-memory",
            dest="data.pin_memory",
            action="store_true",
            help="Pin data to GPU memory for faster data loading",
        )

        parser.add_argument(
            "--drop-last",
            dest="data.drop_last",
            action="store_true",
            help="Drop last incomplete batch",
        )

        parser.add_argument(
            "--shuffle-eval",
            dest="data.shuffle_eval",
            action="store_true",
            help="Shuffle val & test sets",
        )
        return parser


class PLDataModuleFromCorpus(PLDataModuleFromDatasets):
    accepted_tokenizers: List[str] = ["tokenized", "spacy"] + list(
        ALL_PRETRAINED_CONFIG_ARCHIVE_MAP.keys()
    )

    def __init__(
        self,
        train: List,
        train_labels: Optional[List] = None,
        val: Optional[List] = None,
        val_labels: Optional[List] = None,
        test: Optional[List] = None,
        test_labels: Optional[List] = None,
        val_percent: float = 0.2,
        test_percent: float = 0.2,
        batch_size: int = 64,
        batch_size_eval: int = None,
        seed: int = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        drop_last: bool = False,
        shuffle_eval: bool = False,
        sampler_train: Sampler = None,
        sampler_val: Sampler = None,
        sampler_test: Sampler = None,
        batch_sampler_train: BatchSampler = None,
        batch_sampler_val: BatchSampler = None,
        batch_sampler_test: BatchSampler = None,
        collate_fn: Optional[Callable[..., Any]] = None,
        language_model: bool = False,
        tokenizer: str = "spacy",
        **corpus_args,
    ):
        """Wrap raw corpus in a LightningDataModule

        * This handles the selection of the appropriate corpus class based on the tokenizer argument.
        * If language_model=True it uses the appropriate dataset from slp.data.datasets.
        * Uses the PLDataModuleFromDatasets to split the val and test sets if not provided

        Args:
            train (List): Raw train corpus
            train_labels (Optional[List]): Train labels. Defaults to None.
            val (Optional[List]): Raw validation corpus. Defaults to None.
            val_labels (Optional[List]): Validation labels. Defaults to None.
            test (Optional[List]): Raw test corpus. Defaults to None.
            test_labels (Optional[List]): Test labels. Defaults to None.
            val_percent (float): Percent of train to be used for validation if no validation set is given. Defaults to 0.2.
            test_percent (float): Percent of train to be used for test set if no test set is given. Defaults to 0.2.
            batch_size (int): Training batch size. Defaults to 1.
            batch_size_eval (Optional[int]): Validation and test batch size. Defaults to None.
            seed (Optional[int]): Seed for deterministic run. Defaults to None.
            num_workers (int): Number of workers in the DataLoader. Defaults to 1.
            pin_memory (bool): Pin tensors to GPU memory. Defaults to True.
            drop_last (bool): Drop last incomplete batch. Defaults to False.
            sampler_train (Sampler): Sampler for train loader. Defaults to None.
            sampler_val (Sampler): Sampler for validation loader. Defaults to None.
            sampler_test (Sampler): Sampler for test loader. Defaults to None.
            batch_sampler_train (BatchSampler): Batch sampler for train loader. Defaults to None.
            batch_sampler_val (BatchSampler): Batch sampler for validation loader. Defaults to None.
            batch_sampler_test (BatchSampler): Batch sampler for test loader. Defaults to None.
            shuffle_eval (bool): Shuffle validation and test dataloaders. Defaults to False.
            collate_fn (Callable[..., Any]): Collator function. Defaults to None.
            language_model (bool): Use corpus for Language Modeling. Defaults to False.
            tokenizer (str): Select one of the cls.accepted_tokenizers. Defaults to "spacy".

        Raises:
            ValueError: [description]
            ValueError: [description]
        """
        self.language_model = language_model
        if not language_model and train_labels is None:
            raise ValueError(
                "You should provide train labels if not performing language modeling"
            )

        if language_model:
            train_labels = train[1:]
            train = train[0:-1]
            if val is not None:
                val_labels = val[1:]
                val = val[0:-1]
            if test is not None:
                test_labels = test[1:]
                test = test[0:-1]

        train_data = (
            list(zip(train, train_labels)) if train_labels is not None else train
        )
        val_data = None
        if val is not None:
            val_data = list(zip(val, val_labels)) if val_labels is not None else val
        test_data = None
        if test is not None:
            test_data = (
                list(zip(test, test_labels)) if test_labels is not None else test
            )

        super(PLDataModuleFromCorpus, self).__init__(
            train_data,  # type: ignore
            val=val_data,  # type: ignore
            test=test_data,  # type: ignore
            val_percent=val_percent,
            test_percent=test_percent,
            batch_size=batch_size,
            batch_size_eval=batch_size_eval,
            seed=seed,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            shuffle_eval=shuffle_eval,
            sampler_train=sampler_train,
            sampler_val=sampler_val,
            sampler_test=sampler_test,
            batch_sampler_train=batch_sampler_train,
            batch_sampler_val=batch_sampler_val,
            batch_sampler_test=batch_sampler_test,
            collate_fn=collate_fn,
        )

        train_corpus, train_labels = zip(*self.train)  # type: ignore
        val_corpus, val_labels = zip(*self.val)  # type: ignore
        test_corpus, test_labels = zip(*self.test)  # type: ignore

        if tokenizer not in self.accepted_tokenizers:
            raise ValueError(
                f"tokenizer kwarg in {self.__class__.__name__} should be one of {self.accepted_tokenizers}"
            )

        if tokenizer == "spacy":
            logger.info('Selecting WordCorpus because tokenizer="spacy" was provided')
            corpus_cls = WordCorpus  # type: ignore
        elif tokenizer == "tokenized":
            logger.info(
                'Selecting TokenizedCorpus because tokenizer="tokenized" was provided'
            )
            corpus_cls = TokenizedCorpus  # type: ignore
        else:
            logger.info(
                "Selecting HfCorpus because a huggingface tokenizer was provided"
            )
            corpus_cls = HfCorpus  # type: ignore
            corpus_args["tokenizer_model"] = tokenizer

        self.train_corpus = corpus_cls(train_corpus, **corpus_args)  # type: ignore

        if tokenizer == "spacy" or tokenizer == "tokenized":
            # Force train vocabulary on val & test
            corpus_args["word2idx"] = self.train_corpus.word2idx

            if tokenizer == "spacy":
                corpus_args["embeddings"] = self.train_corpus.embeddings
                corpus_args["idx2word"] = self.train_corpus.word2idx

            logger.info(
                "Forcing vocabulary from training set for validation and test sets."
            )

        self.val_corpus = corpus_cls(val_corpus, **corpus_args)  # type: ignore
        self.test_corpus = corpus_cls(test_corpus, **corpus_args)  # type: ignore

        to_tensor = ToTensor(device="cpu")

        if self.language_model:
            self.train = CorpusLMDataset(self.train_corpus).map(to_tensor)
            self.val = CorpusLMDataset(self.val_corpus).map(to_tensor)
            self.test = CorpusLMDataset(self.test_corpus).map(to_tensor)
        else:
            self.train = CorpusDataset(self.train_corpus, train_labels).map(to_tensor)
            self.val = CorpusDataset(self.val_corpus, val_labels).map(to_tensor)
            self.test = CorpusDataset(self.test_corpus, test_labels).map(to_tensor)

    @property
    def embeddings(self) -> Optional[np.ndarray]:
        """Embeddings matrix

        Returns:
            Optional[np.ndarray]: Embeddings matrix
        """
        emb: Optional[np.ndarray] = self.train_corpus.embeddings
        return emb

    @property
    def vocab_size(self) -> int:
        """Number of tokens in the vocabulary

        Returns:
            int: Number of tokens in the vocabulary
        """
        vsz: int = self.train_corpus.vocab_size
        return vsz

    @classmethod
    def add_argparse_args(cls, parent_parser):
        """Augment input parser with arguments for data loading and corpus processing

        Args:
            parent_parser (argparse.ArgumentParser): Parser created by the user

        Returns:
            argparse.ArgumentParser: Augmented parser
        """
        parser = super(PLDataModuleFromCorpus, cls).add_argparse_args(parent_parser)
        parser.add_argument(
            "--tokenizer",
            dest="data.tokenizer",
            type=str.lower,
            # Corpus can already be tokenized, you can use spacy for word tokenization or any tokenizer from hugging face
            choices=cls.accepted_tokenizers,
            help="Token type. The tokenization will happen at this level.",
        )

        # Only when tokenizer == spacy
        parser.add_argument(
            "--limit-vocab",
            dest="data.limit_vocab_size",
            type=int,
            default=-1,
            help="Limit vocab size. -1 means use the whole vocab. Applicable only when --tokenizer=spacy",
        )

        parser.add_argument(
            "--embeddings-file",
            dest="data.embeddings_file",
            type=dir_path,
            help="Path to file with pretrained embeddings. Applicable only when --tokenizer=spacy",
        )

        parser.add_argument(
            "--embeddings-dim",
            dest="data.embeddings_dim",
            type=int,
            help="Embedding dim of pretrained embeddings. Applicable only when --tokenizer=spacy",
        )

        parser.add_argument(
            "--lang",
            dest="data.lang",
            type=str,
            default="en_core_web_md",
            help="Language for spacy tokenizer, e.g. en_core_web_md. Applicable only when --tokenizer=spacy",
        )

        parser.add_argument(
            "--add-specials",
            dest="data.add_special_tokens",
            action="store_true",
            help="Add special tokens for hugging face tokenizers",
        )

        # Generic args
        parser.add_argument(
            "--lower",
            dest="data.lower",
            action="store_true",
            help="Convert to lowercase.",
        )

        parser.add_argument(
            "--prepend-bos",
            dest="data.prepend_bos",
            action="store_true",
            help="Prepend [BOS] token",
        )

        parser.add_argument(
            "--append-eos",
            dest="data.append_eos",
            action="store_true",
            help="Append [EOS] token",
        )

        parser.add_argument(
            "--max-sentence-length",
            dest="data.max_len",
            type=int,
            default=-1,
            help="Maximum allowed sentence length. -1 means use the whole sentence",
        )

        return parser



from enum import Enum


class SPECIAL_TOKENS(Enum):
    """SPECIAL_TOKENS Special Tokens for NLP applications

    Default special tokens values and indices (compatible with BERT):

        * [PAD]: 0
        * [MASK]: 1
        * [UNK]: 2
        * [BOS]: 3
        * [EOS]: 4
        * [CLS]: 5
    """

    PAD = "[PAD]"
    MASK = "[MASK]"
    UNK = "[UNK]"
    BOS = "[BOS]"
    EOS = "[EOS]"
    CLS = "[CLS]"
    SEP = "[SEP]"

    @classmethod
    def has_token(cls, token):
        """has_token check if token exists in SPECIAL_TOKENS

        Args:
            token (str): The special token value (e.g. [PAD]), or name (e.g. PAD)

        Returns:
            bool: True if token exists, False if not
        """
        return any(token == t.name or token == t.value for t in cls)

    @classmethod
    def to_list(cls):
        """to_list Convert SPECIAL_TOKENS to list of tokens

        Returns:
            List[str]: list of token values (e.g. ["[PAD]", "[MASK]"...])
        """
        return list(map(lambda x: x.value, cls))



import torch
from torch import nn
from torch.autograd import Variable


class GaussianNoise(nn.Module):
    def __init__(self, stddev: float, mean: float = 0.0):
        """Additive Gaussian Noise layer

        Args:
            stddev (float): the standard deviation of the distribution
            mean (float): the mean of the distribution
        """
        super().__init__()
        self.stddev = stddev
        self.mean = mean

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Gaussian noise forward pass

        Args:
            x (torch.Tensor): Input features.

        Returns:
            [type]: [description]
        """
        if self.training:
            noise = Variable(x.data.new(x.size()).normal_(self.mean, self.stddev))
            return x + noise
        return x

    def __repr__(self):
        """String representation of class"""
        return "{} (mean={}, stddev={})".format(
            self.__class__.__name__, str(self.mean), str(self.stddev)
        )



from typing import List

from sklearn.preprocessing import LabelEncoder
from torch.utils.data import Dataset


class CorpusLMDataset(Dataset):
    def __init__(self, corpus):
        """Wraps a wikitext dataset from pytorch NLP which is provided as a list of tokens

        Args:
            corpus (List[str] or WordCorpus): List of tokens
        """
        self.source = corpus[:-1]
        self.target = corpus[1:]
        self.transforms = []

    def map(self, t):
        """Append a transform to self.transforms, in order to be applied to the data

        Args:
            t (Callable[[str], Any]): Transform of input token

        Returns:
            CorpusLMDataset: self
        """
        self.transforms.append(t)
        return self

    def __len__(self):
        """Length of corpus

        Returns:
            int: Corpus Length
        """
        return int(len(self.source))

    def __getitem__(self, idx):
        """Get a source and target token from the corpus

        Args:
            idx (int): Token position

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: source=coprus[idx], target=corpus[idx+1]
        """
        src, tgt = self.source[idx], self.target[idx]
        for t in self.transforms:
            src = t(src)
            tgt = t(tgt)
        return src, tgt


class CorpusDataset(Dataset):
    def __init__(self, corpus, labels):
        """Labeled corpus dataset

        Args:
            corpus (WordCorpus, HfCorpus etc..): Input corpus
            labels (List[Any]): Labels for examples
        """
        self.corpus = corpus
        self.labels = labels
        assert len(self.labels) == len(self.corpus), "Incompatible labels and corpus"
        self.transforms = []
        self.label_encoder = None
        if isinstance(self.labels[0], str):
            self.label_encoder = LabelEncoder().fit(self.labels)

    def map(self, t):
        """Append a transform to self.transforms, in order to be applied to the data

        Args:
            t (Callable[[str], Any]): Transform of input token

        Returns:
            CorpusDataset: self
        """
        self.transforms.append(t)
        return self

    def __len__(self):
        """Length of corpus

        Returns:
            int: Corpus Length
        """
        return len(self.corpus)

    def __getitem__(self, idx):
        """Get a source and target token from the corpus

        Args:
            idx (int): Token position

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (processed sentence, label)
        """
        text, target = self.corpus[idx], self.labels[idx]
        if self.label_encoder is not None:
            target = self.label_encoder.transform([target])[0]
        for t in self.transforms:
            text = t(text)
        return text, target



import copy
from typing import Callable, List, Optional, Tuple, Union, cast

import torch
import torch.nn as nn
from loguru import logger
from slp.util import system, types
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class PadPackedSequence(nn.Module):
    def __init__(self, batch_first: bool = True):
        """Wrap sequence padding in nn.Module

        Args:
            batch_first (bool, optional): Use batch first representation. Defaults to True.
        """
        super(PadPackedSequence, self).__init__()
        self.batch_first = batch_first

    def forward(
        self, x: torch.nn.utils.rnn.PackedSequence, lengths: torch.Tensor
    ) -> torch.Tensor:
        """Convert packed sequence to padded sequence

        Args:
            x (torch.nn.utils.rnn.PackedSequence): Packed sequence
            lengths (torch.Tensor): Sorted original sequence lengths

        Returns:
            torch.Tensor: Padded sequence
        """
        max_length = lengths.max().item()
        out, _ = pad_packed_sequence(
            x, batch_first=self.batch_first, total_length=max_length  # type: ignore
        )
        return out  # type: ignore


class PackSequence(nn.Module):
    def __init__(self, batch_first: bool = True):
        """Wrap sequence packing in nn.Module

        Args:
            batch_first (bool, optional): Use batch first representation. Defaults to True.
        """
        super(PackSequence, self).__init__()
        self.batch_first = batch_first

    def forward(
        self, x: torch.Tensor, lengths: torch.Tensor
    ) -> Tuple[torch.nn.utils.rnn.PackedSequence, torch.Tensor]:
        """Pack a padded sequence and sort lengths

        Args:
            x (torch.Tensor): Padded tensor
            lengths (torch.Tensor): Original lengths befor padding

        Returns:
            Tuple[torch.nn.utils.rnn.PackedSequence, torch.Tensor]: (packed sequence, sorted lengths)
        """
        out: torch.nn.utils.rnn.PackedSequence = pack_padded_sequence(
            x, lengths, batch_first=self.batch_first, enforce_sorted=False
        )
        lengths = lengths[out.sorted_indices]
        return out, lengths


def repeat_layer(l: nn.Module, times: int) -> List[nn.Module]:
    """Clone a layer multiple times

    Args:
        l (nn.Module): nn.Module to stack
        times (int): Times to clone

    Returns:
        List[nn.Module]: List of identical clones of input layer
    """
    return [l] + [copy.deepcopy(l) for _ in range(times - 1)]


def pad_mask(
    lengths: torch.Tensor, max_length: Optional[Union[torch.Tensor, int]] = None
) -> torch.Tensor:
    """Generate mask for padded tokens

    Args:
        lengths (torch.Tensor): Original sequence lengths before padding
        max_length (Optional[Union[torch.Tensor, int]], optional): Maximum sequence length. Defaults to None.

    Returns:
        torch.Tensor: padding mask
    """
    if max_length is None:
        max_length = cast(int, torch.max(lengths).item())
    max_length = cast(int, max_length)
    idx = torch.arange(0, max_length, device=lengths.device).unsqueeze(0)
    mask: torch.Tensor = (idx < lengths.unsqueeze(1)).float()
    return mask


def subsequent_mask(max_length: int) -> torch.Tensor:
    """Generate subsequent (lower triangular) mask for transformer autoregressive tasks

    Args:
        max_length (int): Maximum sequence length

    Returns:
        torch.Tensor: The subsequent mask
    """
    mask = torch.ones(max_length, max_length)
    # Ignore typecheck because pytorch types are incomplete
    return mask.triu().t().unsqueeze(0).contiguous()  # type: ignore


def sort_sequences(
    inputs: torch.Tensor, lengths: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor, Callable[[torch.Tensor], torch.Tensor]]:
    """Sort sequences according to lengths (descending)

    Args:
        inputs (torch.Tensor): input sequences, size [B, T, D]
        lengths (torch.Tensor): length of each sequence, size [B]

    Returns:
        Tuple[torch.Tensor, torch.Tensor, Callable[[torch.Tensor], torch.tensor]]:
            (sorted inputs, sorted lengths, function to revert inputs and lengths to unsorted state)
    """
    lengths_sorted, sorted_idx = lengths.sort(descending=True)
    _, unsorted_idx = sorted_idx.sort()

    def unsort(t: torch.Tensor) -> torch.Tensor:
        """Restore original unsorted sequence"""
        return t[unsorted_idx]

    return inputs[sorted_idx], lengths_sorted, unsort


def to_device(
    tt: torch.Tensor, device: Optional[types.Device] = "cpu", non_blocking: bool = False
) -> torch.Tensor:
    """Send a tensor to a device

    Args:
        tt (torch.Tensor): input tensor
        device (Optional[types.Device], optional): Output device. Defaults to "cpu".
        non_blocking (bool, optional): Use blocking or non-blocking memory transfer. Defaults to False.

    Returns:
        torch.Tensor: Tensor in the desired device
    """
    return tt.to(device, non_blocking=non_blocking)


def t_(
    data: types.NdTensor,
    dtype: torch.dtype = torch.float,
    device: Optional[types.Device] = "cpu",
    requires_grad: bool = False,
) -> torch.Tensor:
    """Convert a list or numpy array to torch tensor. If a torch tensor
    is passed it is cast to  dtype, device and the requires_grad flag is
    set IN PLACE.

    Args:
        data: (list, np.ndarray, torch.Tensor): Data to be converted to
            torch tensor.
        dtype: (torch.dtype): The type of the tensor elements
            (Default value = torch.float)
        device: (torch.device, str): Device where the tensor should be
            (Default value = 'cpu')
        requires_grad: bool): Trainable tensor or not? (Default value = False)

    Returns:
        (torch.Tensor): A tensor of appropriate dtype, device and
            requires_grad containing data

    """
    if isinstance(device, str):
        device = torch.device(device)

    tt = torch.as_tensor(data, dtype=dtype, device=device).requires_grad_(requires_grad)
    return tt


def t(
    data: types.NdTensor,
    dtype: torch.dtype = torch.float,
    device: types.Device = "cpu",
    requires_grad: bool = False,
) -> torch.Tensor:
    """Convert a list or numpy array to torch tensor. If a torch tensor
    is passed it is cast to  dtype, device and the requires_grad flag is
    set. This always copies data.

    Args:
        data: (list, np.ndarray, torch.Tensor): Data to be converted to
            torch tensor.
        dtype: (torch.dtype): The type of the tensor elements
            (Default value = torch.float)
        device: (torch.device, str): Device where the tensor should be
            (Default value = 'cpu')
        requires_grad: (bool): Trainable tensor or not? (Default value = False)

    Returns:
        (torch.Tensor): A tensor of appropriate dtype, device and
            requires_grad containing data

    """
    tt = torch.tensor(data, dtype=dtype, device=device, requires_grad=requires_grad)
    return tt


def mktensor(
    data: types.NdTensor,
    dtype: torch.dtype = torch.float,
    device: types.Device = "cpu",
    requires_grad: bool = False,
    copy: bool = True,
) -> torch.Tensor:
    """Convert a list or numpy array to torch tensor. If a torch tensor
        is passed it is cast to  dtype, device and the requires_grad flag is
        set. This can copy data or make the operation in place.

    Args:
        data: (list, np.ndarray, torch.Tensor): Data to be converted to
            torch tensor.
        dtype: (torch.dtype): The type of the tensor elements
            (Default value = torch.float)
        device: (torch.device, str): Device where the tensor should be
            (Default value = 'cpu')
        requires_grad: (bool): Trainable tensor or not? (Default value = False)
        copy: (bool): If false creates the tensor inplace else makes a copy
            (Default value = True)

    Returns:
        (torch.Tensor): A tensor of appropriate dtype, device and
            requires_grad containing data

    """
    tensor_factory = t if copy else t_
    return tensor_factory(data, dtype=dtype, device=device, requires_grad=requires_grad)


def from_checkpoint(
    checkpoint_file: Optional[str],
    obj: types.ModuleOrOptimizer,
    map_location: Optional[types.Device] = "cpu",
    dataparallel: bool = False,
) -> types.ModuleOrOptimizer:
    """Load model or optimizer from saved state_dict

    Args:
        checkpoint_file (Optional[str]): File containing the state dict
        obj (types.ModuleOrOptimizer): Module or optimizer instance to load the checkpoint
        map_location (Optional[types.Device], optional): Where to load. Defaults to "cpu".
        dataparallel (bool, optional): If data parallel remove leading "module." from statedict keys. Defaults to False.

    Returns:
        types.ModuleOrOptimizer: Loaded module or optimizer
    """
    if checkpoint_file is None:
        return obj

    if not system.is_file(checkpoint_file):
        logger.warning(
            f"The checkpoint {checkpoint_file} you are trying to load "
            "does not exist. Continuing without loading..."
        )
        return obj

    state_dict = torch.load(checkpoint_file, map_location=map_location)
    if dataparallel:
        state_dict = {k.replace("module.", ""): v for k, v in state_dict.items()}
    obj.load_state_dict(state_dict)
    return obj


def rotate_tensor(l: torch.Tensor, n: int = 1) -> torch.Tensor:
    """Roate tensor by n positions to the right

    Args:
        l (torch.Tensor): input tensor
        n (int, optional): positions to rotate. Defaults to 1.

    Returns:
        torch.Tensor: rotated tensor
    """
    return torch.cat((l[n:], l[:n]))


def shift_tensor(l: torch.Tensor, n: int = 1) -> torch.Tensor:
    """Shift tensor by n positions

    Args:
        l (torch.Tensor): input tensor
        n (int, optional): positions to shift. Defaults to 1.

    Returns:
        torch.Tensor: shifted tensor
    """
    out = rotate_tensor(l, n=n)
    out[-n:] = 0
    return out



import argparse
import os
from typing import Optional, Sequence, Union

import pytorch_lightning as pl
import torch.nn as nn
from loguru import logger
from slp.plbind.helpers import EarlyStoppingWithLogs, FixedWandbLogger
from slp.util.system import date_fname, has_internet_connection, safe_mkdirs
from slp.util.types import dir_path


def add_trainer_args(parent_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Augment parser with trainer arguments

    Args:
        parent_parser (argparse.ArgumentParser): Parser created by the user

    Returns:
        argparse.ArgumentParser: Augmented parser
    """
    parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument(
        "--seed",
        dest="seed",
        type=int,
        default=None,
        help="Seed for reproducibility",
    )

    parser.add_argument(
        "--config",
        dest="config",
        type=dir_path,
        default=None,
        help="Path to YAML configuration file",
    )

    parser.add_argument(
        "--experiment-name",
        dest="trainer.experiment_name",
        type=str,
        help="Name of the running experiment",
    )

    parser.add_argument(
        "--run-id",
        dest="trainer.run_id",
        type=str,
        help="Unique identifier for the current run. If not provided it is inferred from datetime.now()",
    )

    parser.add_argument(
        "--experiment-group",
        dest="trainer.experiment_group",
        type=str,
        help="Group of current experiment. Useful when evaluating for different seeds / cross-validation etc.",
    )

    parser.add_argument(
        "--experiments-folder",
        dest="trainer.experiments_folder",
        type=str,
        default="experiments",
        help="Top-level folder where experiment results & checkpoints are saved",
    )

    parser.add_argument(
        "--save-top-k",
        dest="trainer.save_top_k",
        type=int,
        help="Save checkpoints for top k models",
    )

    parser.add_argument(
        "--patience",
        dest="trainer.patience",
        type=int,
        help="Number of epochs to wait before early stopping",
    )

    parser.add_argument(
        "--wandb-project",
        dest="trainer.wandb_project",
        type=str,
        help="Wandb project under which results are saved",
    )

    parser.add_argument(
        "--tags",
        dest="trainer.tags",
        type=str,
        nargs="*",
        help="Tags for current run to make results searchable.",
    )

    parser.add_argument(
        "--stochastic_weight_avg",
        dest="trainer.stochastic_weight_avg",
        action="store_true",
        help="Use Stochastic weight averaging.",
    )

    parser.add_argument(
        "--gpus", dest="trainer.gpus", type=int, help="Number of GPUs to use"
    )

    parser.add_argument(
        "--val-interval",
        dest="trainer.check_val_every_n_epoch",
        type=int,
        default=1,
        help="Run validation every n epochs",
    )

    parser.add_argument(
        "--clip-grad-norm",
        dest="trainer.gradient_clip_val",
        type=float,
        help="Clip gradients with ||grad(w)|| >= args.clip_grad_norm",
    )

    parser.add_argument(
        "--epochs",
        dest="trainer.max_epochs",
        type=int,
        help="Maximum number of training epochs",
    )

    parser.add_argument(
        "--steps",
        dest="trainer.max_steps",
        type=int,
        help="Maximum number of training steps",
    )

    parser.add_argument(
        "--tbtt_steps",
        dest="trainer.truncated_bptt_steps",
        type=int,
        help="Truncated Back-propagation-through-time steps.",
    )

    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="If true, we run a full run on a small subset of the input data and overfit 10 training batches",
    )

    parser.add_argument(
        "--offline",
        dest="trainer.force_wandb_offline",
        action="store_true",
        help="If true, forces offline execution of wandb logger",
    )

    parser.add_argument(
        "--early-stop-on",
        dest="trainer.early_stop_on",
        type=str,
        default="val_loss",
        help="Metric for early stopping",
    )

    parser.add_argument(
        "--early-stop-mode",
        dest="trainer.early_stop_mode",
        type=str,
        choices=["min", "max"],
        default="min",
        help="Minimize or maximize early stopping metric",
    )

    return parser


def add_optimizer_args(
    parent_parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Augment parser with optimizer arguments

    Args:
        parent_parser (argparse.ArgumentParser): Parser created by the user

    Returns:
        argparse.ArgumentParser: Augmented parser
    """
    parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument(
        "--optimizer",
        dest="optimizer",
        type=str,
        choices=[
            "Adam",
            "AdamW",
            "SGD",
            "Adadelta",
            "Adagrad",
            "Adamax",
            "ASGD",
            "RMSprop",
        ],
        default="Adam",
        help="Which optimizer to use",
    )

    parser.add_argument(
        "--lr",
        dest="optim.lr",
        type=float,
        default=1e-3,
        help="Learning rate",
    )

    parser.add_argument(
        "--weight-decay",
        dest="optim.weight_decay",
        type=float,
        help="Learning rate",
    )

    parser.add_argument(
        "--lr-scheduler",
        dest="lr_scheduler",
        action="store_true",
        # type=str,
        # choices=["ReduceLROnPlateau"],
        help="Use learning rate scheduling. Currently only ReduceLROnPlateau is supported out of the box",
    )

    parser.add_argument(
        "--lr-factor",
        dest="lr_schedule.factor",
        type=float,
        help="Multiplicative factor by which LR is reduced. Used if --lr-scheduler is provided.",
    )

    parser.add_argument(
        "--lr-patience",
        dest="lr_schedule.patience",
        type=int,
        help="Number of epochs with no improvement after which learning rate will be reduced. Used if --lr-scheduler is provided.",
    )

    parser.add_argument(
        "--lr-cooldown",
        dest="lr_schedule.cooldown",
        type=int,
        help="Number of epochs to wait before resuming normal operation after lr has been reduced. Used if --lr-scheduler is provided.",
    )

    parser.add_argument(
        "--min-lr",
        dest="lr_schedule.min_lr",
        type=float,
        help="Minimum lr for LR scheduling. Used if --lr-scheduler is provided.",
    )

    return parser


def make_trainer(
    experiment_name: str = "experiment",
    experiment_description: Optional[str] = None,
    run_id: Optional[str] = None,
    experiment_group: Optional[str] = None,
    experiments_folder: str = "experiments",
    save_top_k: int = 3,
    patience: int = 3,
    wandb_project: Optional[str] = None,
    wandb_user: Optional[str] = None,
    force_wandb_offline: bool = False,
    tags: Optional[Sequence] = None,
    stochastic_weight_avg: bool = False,
    auto_scale_batch_size: bool = False,
    gpus: int = 0,
    check_val_every_n_epoch: int = 1,
    gradient_clip_val: float = 0,
    precision: int = 32,
    max_epochs: Optional[int] = 100,
    max_steps: Optional[int] = None,
    truncated_bptt_steps: Optional[int] = None,
    fast_dev_run: Optional[int] = None,
    overfit_batches: Optional[int] = None,
    terminate_on_nan: bool = False,  # Be careful this makes training very slow for large models
    profiler: Optional[Union[pl.profiler.BaseProfiler, bool, str]] = "simple",
    early_stop_on: str = "val_loss",
    early_stop_mode: str = "min",
) -> pl.Trainer:
    """Configure trainer with preferred defaults

    * Experiment folder and run_id configured (based on datetime.now())
    * Wandb and CSV loggers run by default
    * Wandb configured to save code and checkpoints
    * Wandb configured in online mode except if no internet connection is available
    * Early stopping on best validation loss is configured by default
    * Checkpointing on best validation loss is configured by default
    *

    Args:
        experiment_name (str, optional): Experiment name. Defaults to "experiment".
        experiment_description (Optional[str], optional): Detailed description of the experiment. Defaults to None.
        run_id (Optional[str], optional): Unique run_id. Defaults to datetime.now(). Defaults to None.
        experiment_group (Optional[str], optional): Group experiments over multiple runs. Defaults to None.
        experiments_folder (str, optional): Folder to save outputs. Defaults to "experiments".
        save_top_k (int, optional): Save top k checkpoints. Defaults to 3.
        patience (int, optional): Patience for early stopping. Defaults to 3.
        wandb_project (Optional[str], optional): Wandb project to save the experiment. Defaults to None.
        wandb_user (Optional[str], optional): Wandb username. Defaults to None.
        force_wandb_offline (bool): Force offline execution of wandb
        tags (Optional[Sequence], optional): Additional tags to attach to the experiment. Defaults to None.
        stochastic_weight_avg (bool, optional): Use stochastic weight averaging. Defaults to False.
        auto_scale_batch_size (bool, optional): Find optimal batch size for the available resources when running
                trainer.tune(). Defaults to False.
        gpus (int, optional): number of GPUs to use. Defaults to 0.
        check_val_every_n_epoch (int, optional): Run validation every n epochs. Defaults to 1.
        gradient_clip_val (float, optional): Clip gradient norm value. Defaults to 0 (no clipping).
        precision (int, optional): Floating point precision. Defaults to 32.
        max_epochs (Optional[int], optional): Maximum number of epochs for training. Defaults to 100.
        max_steps (Optional[int], optional): Maximum number of steps for training. Defaults to None.
        truncated_bptt_steps (Optional[int], optional): Truncated back prop breaks performs backprop every k steps of much longer
                sequence. Defaults to None.
        fast_dev_run (Optional[int], optional): Run training on a small number of  batches for debugging. Defaults to None.
        overfit_batches (Optional[int], optional): Try to overfit a small number of batches for debugging. Defaults to None.
        terminate_on_nan (bool, optional): Terminate on NaN gradients. Warning this makes training slow. Defaults to False.
        profiler (Optional[Union[pl.profiler.BaseProfiler, bool, str]]): Use profiler to track execution times of each function
        early_stop_on (str): metric for early stopping
        early_stop_mode (str): "min" or "max"

    Returns:
        pl.Trainer: Configured trainer
    """
    if overfit_batches is not None:
        trainer = pl.Trainer(overfit_batches=overfit_batches, gpus=gpus)
        return trainer

    if fast_dev_run is not None:
        trainer = pl.Trainer(fast_dev_run=fast_dev_run, gpus=gpus)
        return trainer

    logging_dir = os.path.join(experiments_folder, experiment_name)
    safe_mkdirs(logging_dir)

    run_id = run_id if run_id is not None else date_fname()
    if run_id is None:
        run_id = date_fname()

    if run_id in os.listdir(logging_dir):
        logger.warning(
            "The run id you provided {run_id} already exists in {logging_dir}"
        )
        run_id = date_fname()
        logger.info("Setting run_id={run_id}")

    checkpoint_dir = os.path.join(logging_dir, run_id, "checkpoints")

    logger.info(f"Logs will be saved in {logging_dir}")
    logger.info(f"Logs will be saved in {checkpoint_dir}")

    if wandb_project is None:
        wandb_project = experiment_name

    connected = has_internet_connection()
    offline_run = force_wandb_offline or not connected

    loggers = [
        pl.loggers.CSVLogger(logging_dir, name="csv_logs", version=run_id),
        FixedWandbLogger(
            name=experiment_name,
            project=wandb_project,
            anonymous=False,
            save_dir=logging_dir,
            version=run_id,
            save_code=True,
            checkpoint_dir=checkpoint_dir,
            offline=offline_run,
            log_model=not offline_run,
            entity=wandb_user,
            group=experiment_group,
            notes=experiment_description,
            tags=tags,
        ),
    ]

    logger.info(f"Configured wandb and CSV loggers.")
    logger.info(
        f"Wandb configured to run {experiment_name}/{run_id} in project {wandb_project}"
    )
    if connected:
        logger.info(f"Results will be stored online.")
    else:
        logger.info(f"Results will be stored offline due to bad internet connection.")
        logger.info(
            f"If you want to upload your results later run\n\t wandb sync {logging_dir}/wandb/run-{run_id}"
        )

    if experiment_description is not None:
        logger.info(
            f"Experiment verbose description:\n{experiment_description}\n\nTags:{'n/a' if tags is None else tags}"
        )

    callbacks = [
        EarlyStoppingWithLogs(
            monitor=early_stop_on,
            mode=early_stop_mode,
            patience=patience,
            verbose=True,
        ),
        pl.callbacks.ModelCheckpoint(
            dirpath=checkpoint_dir,
            filename="{epoch}-{val_loss:.2f}",
            monitor=early_stop_on,
            save_top_k=save_top_k,
            mode=early_stop_mode,
        ),
        pl.callbacks.LearningRateMonitor(logging_interval="step"),
    ]

    logger.info("Configured Early stopping and Model checkpointing to track val_loss")

    trainer = pl.Trainer(
        default_root_dir=logging_dir,
        gpus=gpus,
        max_epochs=max_epochs,
        max_steps=max_steps,
        callbacks=callbacks,
        logger=loggers,
        check_val_every_n_epoch=check_val_every_n_epoch,
        gradient_clip_val=gradient_clip_val,
        auto_scale_batch_size=auto_scale_batch_size,
        stochastic_weight_avg=stochastic_weight_avg,
        precision=precision,
        truncated_bptt_steps=truncated_bptt_steps,
        terminate_on_nan=terminate_on_nan,
        progress_bar_refresh_rate=10,
        profiler=profiler,
    )

    return trainer


def watch_model(trainer: pl.Trainer, model: nn.Module) -> None:
    """If wandb logger is configured track gradient and weight norms

    Args:
        trainer (pl.Trainer): Trainer
        model (nn.Module): Module to watch
    """
    if isinstance(trainer.logger.experiment, list):
        for log in trainer.logger.experiment:
            try:
                log.watch(model, log="all")
                logger.info("Tracking model weights & gradients in wandb.")
                break
            except:
                pass
    else:
        try:
            trainer.logger.experiment.watch(model, log="all")
            logger.info("Tracking model weights & gradients in wandb.")
        except:
            pass



import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from slp.plbind.dm import PLDataModuleFromDatasets
from slp.plbind.module import AutoEncoderPLModule
from slp.plbind.trainer import make_trainer, watch_model
from slp.util.log import configure_logging
from torch.optim import Adam
from torchvision.datasets import MNIST  # type: ignore
from torchvision.transforms import Compose, Normalize, ToTensor  # type: ignore


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 16, 3, padding=1)
        self.conv2 = nn.Conv2d(16, 4, 3, padding=1)
        self.pool = nn.MaxPool2d(2, 2)

        self.t_conv1 = nn.ConvTranspose2d(4, 16, 2, stride=2)
        self.t_conv2 = nn.ConvTranspose2d(16, 1, 2, stride=2)

    def forward(self, x):
        # encode
        x = F.relu(self.conv1(x))
        x = self.pool(x)
        x = F.relu(self.conv2(x))
        x = self.pool(x)  # compressed representation

        # decode
        x = F.relu(self.t_conv1(x))
        x = F.sigmoid(self.t_conv2(x))

        return x


def get_data():
    data_transform = Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])
    train = MNIST(download=True, root=".", transform=data_transform, train=True)

    val = MNIST(download=False, root=".", transform=data_transform, train=False)
    return train, val


if __name__ == "__main__":

    EXPERIMENT_NAME = "mnist-autoencoder"

    configure_logging(f"logs/{EXPERIMENT_NAME}")

    train, test = get_data()

    ldm = PLDataModuleFromDatasets(
        train, test=test, batch_size=128, batch_size_eval=256
    )

    model = Net()
    optimizer = Adam(model.parameters(), lr=1e-3)
    criterion = nn.MSELoss()

    lm = AutoEncoderPLModule(model, optimizer, criterion)

    trainer = make_trainer(
        EXPERIMENT_NAME, max_epochs=5, gpus=1, wandb_project="testpl"
    )
    watch_model(trainer, model)

    trainer.fit(lm, datamodule=ldm)

    trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())



from slp.data.collators import Seq2SeqCollator, SequenceClassificationCollator
from slp.data.corpus import HfCorpus, WordCorpus, create_vocab
from slp.data.datasets import CorpusDataset, CorpusLMDataset
from slp.data.transforms import (
    HuggingFaceTokenizer,
    ReplaceUnknownToken,
    SentencepieceTokenizer,
    SpacyTokenizer,
    ToTensor,
    ToTokenIds,
)



import torch
import torch.nn as nn


class Classifier(nn.Module):
    def __init__(
        self,
        encoder: nn.Module,
        encoded_features: int,
        num_classes: int,
        dropout: float = 0.2,
    ):
        """Classifier wrapper module

        Stores a Neural Network encoder and adds a classification layer on top.

        Args:
            encoder (nn.Module): [description]
            encoded_features (int): [description]
            num_classes (int): [description]
            dropout (float): Drop probability
        """
        super(Classifier, self).__init__()
        self.encoder = encoder
        self.drop = nn.Dropout(dropout)
        self.clf = nn.Linear(encoded_features, num_classes)

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """Encode inputs using the encoder network and perform classification

        Returns:
            torch.Tensor: [B, *, num_classes] Logits tensor
        """
        encoded: torch.Tensor = self.encoder(*args, **kwargs)  # type: ignore
        out: torch.Tensor = self.drop(encoded)
        out = self.clf(out)
        return out



import math
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from loguru import logger
from slp.modules.regularization import GaussianNoise


class PositionalEncoding(nn.Module):
    def __init__(self, embedding_dim: int = 512, max_len: int = 5000):
        """Inject some information about the relative or absolute position of the tokens in the sequence.

        The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.

        PE for even positions:

        $$\\text{PosEncoder}(pos, 2i) = sin(\\frac{pos}{10000^{\\frac{2i}{d}}})$$

        PE for odd positions:

        $$\\text{PosEncoder}(pos, 2i+1) = cos(\\frac{pos}{10000^{\\frac{2i}{d}}})$$

        where $pos$ is the word position and $i$ is the embedding idx

        Implementation modified from pytorch/examples/word_language_model.py

        Args:
            embedding_dim (int): Embedding / model dimension. Defaults to 512.
            max_len (int): Maximum sequence length that can be encoded. Defaults to 5000.
        """
        super(PositionalEncoding, self).__init__()
        pe = torch.zeros(max_len, embedding_dim)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, embedding_dim, 2).float()
            * (-math.log(10000.0) / embedding_dim)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Calculate positional embeddings for input and add them to input tensor

        $$out = x + PosEmbed(x)$$

        x is assumed to be batch first

        Args:
            x (torch.Tensor): [B, L, D] input embeddings

        Returns:
            torch.Tensor: Embeddings + positional embeddings
        """
        x = x + self.pe[:, : x.size(1), :]  # type: ignore
        return x


class Embed(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        embeddings: Optional[np.ndarray] = None,
        noise: float = 0.0,
        dropout: float = 0.0,
        scale: float = 1.0,
        trainable: bool = False,
    ):
        """
        Define the layer of the model and perform the initializations
        of the layers (wherever it is necessary)

        Args:
            num_embeddings (int): Total number of embeddings.
            embeddings_dim (int): Embedding dimension.
            embeddings (numpy.ndarray): the 2D ndarray with the word vectors.
            noise (float): Optional additive noise. Defaults to 0.0.
            dropout (float): Embedding dropout probability. Defaults to 0.0.
            scale (float): Scale word embeddings by a constant. Defaults to 1.0.
            trainable (bool): Finetune embeddings. Defaults to False
        """
        super(Embed, self).__init__()
        self.scale = scale  # scale embeddings by value. Needed for transformer
        # define the embedding layer, with the corresponding dimensions
        self.embedding = nn.Embedding(
            num_embeddings=num_embeddings, embedding_dim=embedding_dim
        )

        if embeddings is not None:
            logger.info("Initializing Embedding layer with pre-trained weights.")
            if trainable:
                logger.info("Embeddings are going to be finetuned")
            else:
                logger.info("Embeddings are frozen")
            self.init_embeddings(embeddings, trainable)

        # the dropout "layer" for the word embeddings
        self.dropout = nn.Dropout(dropout)

        # the gaussian noise "layer" for the word embeddings
        self.noise = GaussianNoise(noise)

    def init_embeddings(self, weights: np.ndarray, trainable: bool):
        """Initialize embeddings matrix with pretrained embeddings

        Args:
            weights (np.ndarray): pretrained embeddings
            trainable (bool): Finetune embeddings?
        """
        self.embedding.weight = nn.Parameter(
            torch.from_numpy(weights), requires_grad=trainable
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Embed input tokens

        Assign embedding that corresponds to each token.
        Optionally add Gaussian noise and embedding dropout and scale embeddings by a constant.

        Args:
            x (torch.Tensor): [B, L] Input token ids.

        Returns:
            (torch.Tensor) -> [B, L, E] Embedded tokens.
        """
        embeddings = self.embedding(x)

        if self.noise.stddev > 0:
            embeddings = self.noise(embeddings)

        if self.dropout.p > 0:
            embeddings = self.dropout(embeddings)

        return embeddings * self.scale  # type: ignore



# python examples/mnist_rnn.py --bsz 128 --bsz-eval 256

import sys
from argparse import ArgumentParser

import pytorch_lightning as pl
import torch.nn as nn
import torch.optim as optim
from loguru import logger
from slp.config.config_parser import make_cli_parser, parse_config
from slp.data.collators import SequenceClassificationCollator
from slp.modules.rnn import RNN
from slp.plbind import (
    FromLogits,
    PLDataModuleFromDatasets,
    RnnPLModule,
    make_trainer,
    watch_model,
)
from slp.util.log import configure_logging
from torchvision.datasets import MNIST  # type: ignore
from torchvision.transforms import Compose, Normalize, ToTensor  # type: ignore

collate_fn = SequenceClassificationCollator()


class Net(nn.Module):
    def __init__(self, input_size, hidden_size=40, num_classes=10, bidirectional=False):
        super().__init__()
        self.encoder = RNN(input_size, hidden_size, bidirectional=bidirectional)
        out_size = hidden_size if not bidirectional else 2 * hidden_size
        self.clf = nn.Linear(out_size, num_classes)

    def forward(self, x, lengths):
        _, x, _ = self.encoder(x, lengths)
        out = self.clf(x)
        return out


def get_parser():
    parser = ArgumentParser("MNIST classification example")
    parser.add_argument(
        "--hidden",
        dest="model.hidden_size",
        type=int,
        help="Intermediate hidden layers for linear module",
    )
    parser.add_argument(
        "--bi",
        dest="model.bidirectional",
        action="store_true",
        help="Use BiLSTM",
    )
    return parser


def get_data():
    def squeeze(x):
        return x.squeeze()

    data_transform = Compose([ToTensor(), Normalize((0.1307,), (0.3081,)), squeeze])
    train = MNIST(download=True, root=".", transform=data_transform, train=True)

    val = MNIST(download=False, root=".", transform=data_transform, train=False)
    return train, val


if __name__ == "__main__":
    # SETUP ##################################################
    parser = get_parser()
    parser = make_cli_parser(parser, PLDataModuleFromDatasets)

    config = parse_config(parser, parser.parse_args().config)

    if config.trainer.experiment_name == "experiment":
        config.trainer.experiment_name = "mnist-rnn-classification"

    configure_logging(f"logs/{config.trainer.experiment_name}")

    if config.seed is not None:
        logger.info("Seeding everything with seed={seed}")
        pl.utilities.seed.seed_everything(seed=config.seed)

    train, test = get_data()

    # Get data and make datamodule ##########################
    ldm = PLDataModuleFromDatasets(
        train, test=test, seed=config.seed, collate_fn=collate_fn, **config.data
    )

    # Create model, optimizer, criterion, scheduler ###########
    model = Net(28, **config.model)

    optimizer = getattr(optim, config.optimizer)(model.parameters(), **config.optim)
    criterion = nn.CrossEntropyLoss()

    lr_scheduler = None
    if config.lr_scheduler:
        lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, **config.lr_schedule
        )

    # Wrap in PLModule, & configure metrics ####################
    lm = RnnPLModule(
        model,
        optimizer,
        criterion,
        lr_scheduler=lr_scheduler,
        metrics={"acc": FromLogits(pl.metrics.classification.Accuracy())},
        hparams=config,
    )

    # Run debugging session or fit & test the model ############
    if config.debug:
        logger.info("Running in debug mode: Fast run on 5 batches")
        trainer = make_trainer(fast_dev_run=5)
        trainer.fit(lm, datamodule=ldm)

        logger.info("Running in debug mode: Overfitting 5 batches")
        trainer = make_trainer(overfit_batches=5)
        trainer.fit(lm, datamodule=ldm)

    else:
        trainer = make_trainer(**config.trainer)
        watch_model(trainer, model)

        trainer.fit(lm, datamodule=ldm)

        trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())

        logger.info("Run finished. Uploading files to wandb...")



import os
from typing import Any, Dict, List, Optional

import sentencepiece as spm
import spacy
import torch
from slp.config.nlp import SPECIAL_TOKENS
from slp.util.pytorch import mktensor
from spacy.attrs import ORTH
from transformers import AutoTokenizer

# Avoid deadlocks for hugging face tokenizers
os.environ["TOKENIZERS_PARALLELISM"] = "false"


class SentencepieceTokenizer(object):
    def __init__(
        self,
        lower: bool = True,
        model: Optional[Any] = None,
        prepend_bos: bool = False,
        append_eos: bool = False,
        specials: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
    ):
        """Tokenize sentence using pretrained sentencepiece model

        Args:
            lower (bool): Lowercase string. Defaults to True.
            model (Optional[Any]): Sentencepiece model. Defaults to None.
            prepend_bos (bool): Prepend BOS for seq2seq. Defaults to False.
            append_eos (bool): Append EOS for seq2seq. Defaults to False.
            specials (Optional[SPECIAL_TOKENS]): Special tokens. Defaults to SPECIAL_TOKENS.
        """
        self.tokenizer = spm.SentencePieceProcessor()
        self.tokenizer.Load(model)
        self.specials = specials
        self.lower = lower
        self.vocab_size = self.tokenizer.get_piece_size()
        self.pre_id = []
        self.post_id = []
        if prepend_bos:
            self.pre_id.append(self.tokenizer.piece_to_id(self.specials.BOS.value))  # type: ignore
        if append_eos:
            self.post_id.append(self.tokenizer.piece_to_id(self.specials.EOS.value))  # type: ignore

    def __call__(self, x: str) -> List[int]:
        """Call to tokenize function

        Args:
            x (str): Input string

        Returns:
            List[int]: List of tokens ids
        """
        if self.lower:
            x = x.lower()
        ids: List[int] = self.pre_id + self.tokenizer.encode_as_ids(x) + self.post_id
        return ids


class HuggingFaceTokenizer(object):
    def __init__(
        self,
        lower: bool = True,
        model: str = "bert-base-uncased",
        add_special_tokens: bool = True,
    ):
        """Apply one of huggingface tokenizers to a string

        Args:
            lower (bool): Lowercase string. Defaults to True.
            model (str): Select transformer model. Defaults to "bert-base-uncased".
            add_special_tokens (bool): Insert special tokens to tokenized string. Defaults to True.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(model, do_lower_case=lower)
        self.tokenizer.max_len = 65536  # hack to suppress warnings
        self.vocab_size = len(self.tokenizer.vocab)
        self.add_special_tokens = add_special_tokens

    def detokenize(self, x: List[int]) -> List[str]:
        """Convert list of token ids to list of tokens

        Args:
            x (List[int]): List of token ids

        Returns:
            List[str]: List of tokens
        """
        out: List[str] = self.tokenizer.convert_ids_to_tokens(x)
        return out

    def __call__(self, x: str) -> List[int]:
        """Call to tokenize function

        Args:
            x (str): Input string

        Returns:
            List[int]: List of token ids
        """
        out: List[int] = self.tokenizer.encode(
            x, add_special_tokens=self.add_special_tokens
        )
        return out


class SpacyTokenizer(object):
    def __init__(
        self,
        lower: bool = True,
        prepend_bos: bool = False,
        append_eos: bool = False,
        specials: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
        lang: str = "en_core_web_sm",
    ):
        """Apply spacy tokenizer to str

        Args:
            lower (bool): Lowercase string. Defaults to True.
            prepend_bos (bool): Prepend BOS for seq2seq. Defaults to False.
            append_eos (bool): Append EOS for seq2seq. Defaults to False.
            specials (Optional[SPECIAL_TOKENS]): Special tokens. Defaults to SPECIAL_TOKENS.
            lang (str): Spacy language, e.g. el_core_web_sm, en_core_web_sm etc. Defaults to "en_core_web_md".
        """
        self.lower = lower
        self.specials = SPECIAL_TOKENS
        self.lang = lang
        self.pre_id = []
        self.post_id = []
        if prepend_bos:
            self.pre_id.append(self.specials.BOS.value)
        if append_eos:
            self.post_id.append(self.specials.EOS.value)
        self.nlp = self.get_nlp(name=lang, specials=specials)

    def get_nlp(
        self,
        name: str = "en_core_web_sm",
        specials: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
    ) -> spacy.Language:
        """Get spacy nlp object for given lang and add SPECIAL_TOKENS

        Args:
            name (str): Spacy language, e.g. el_core_web_sm, en_core_web_sm etc. Defaults to "en_core_web_md".
            specials (Optional[SPECIAL_TOKENS]): Special tokens. Defaults to SPECIAL_TOKENS.

        Returns:
            spacy.Language: spacy text-processing pipeline
        """
        nlp = spacy.load(name)
        if specials is not None:
            for token in specials.to_list():
                nlp.tokenizer.add_special_case(token, [{ORTH: token}])
        return nlp

    def __call__(self, x: str) -> List[str]:
        """Call to tokenize function

        Args:
            x (str): Input string

        Returns:
            List[str]: List of tokens
        """
        if self.lower:
            x = x.lower()
        out: List[str] = (
            self.pre_id + [y.text for y in self.nlp.tokenizer(x)] + self.post_id
        )
        return out


class ToTokenIds(object):
    def __init__(
        self,
        word2idx: Dict[str, int],
        specials: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
    ):
        """Convert List of tokens to list of token ids

        Args:
            word2idx (Dict[str, int]): Word to index mapping
            specials (Optional[SPECIAL_TOKENS]): Special tokens. Defaults to SPECIAL_TOKENS.
        """
        self.word2idx = word2idx
        self.unk_value = specials.UNK.value if specials is not None else "[UNK]"  # type: ignore

    def __call__(self, x: List[str]) -> List[int]:
        """Convert list of tokens to list of token ids

        Args:
            x (List[str]): List of tokens

        Returns:
            List[int]: List of token ids
        """
        return [
            self.word2idx[w] if w in self.word2idx else self.word2idx[self.unk_value]
            for w in x
        ]


class ReplaceUnknownToken(object):
    def __init__(
        self,
        old_unk: str = "<unk>",
        new_unk: str = SPECIAL_TOKENS.UNK.value,  # type: ignore
    ):
        """Replace existing unknown tokens in the vocab to [UNK]. Useful for wikitext

        Args:
            old_unk (str): Unk token in corpus. Defaults to "<unk>".
            new_unk (str): Desired unk value. Defaults to SPECIAL_TOKENS.UNK.value.
        """
        self.old_unk = old_unk
        self.new_unk = new_unk

    def __call__(self, x: List[str]) -> List[str]:
        """Convert <unk> in list of tokens to [UNK]

        Args:
            x (List[str]): List of tokens

        Returns:
            List[str]: List of tokens
        """
        return [w if w != self.old_unk else self.new_unk for w in x]


class ToTensor(object):
    def __init__(self, device: str = "cpu", dtype: torch.dtype = torch.long):
        """To tensor convertor

        Args:
            device (str): Device to map the tensor. Defaults to "cpu".
            dtype (torch.dtype): Type of resulting tensor. Defaults to torch.long.
        """
        self.device = device
        self.dtype = dtype

    def __call__(self, x: List[Any]) -> torch.Tensor:
        """Convert list of tokens or list of features to tensor

        Args:
            x (List[Any]): List of tokens or features

        Returns:
            torch.Tensor: Resulting tensor
        """
        return mktensor(x, device=self.device, dtype=self.dtype)



import argparse
import os
from typing import Any, Callable, Dict, List, Tuple, TypeVar, Union

import numpy as np
import torch
import validators
from omegaconf import DictConfig
from torch.optim.optimizer import Optimizer

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

NdTensor = Union[np.ndarray, torch.Tensor, List[T]]
Label = Union[NdTensor, int]

Device = Union[torch.device, str]

ModuleOrOptimizer = Union[torch.nn.Module, Optimizer]

LossType = Union[torch.nn.Module, Callable]

# word2idx, idx2word, embedding vectors
Embeddings = Tuple[Dict[str, int], Dict[int, str], np.ndarray]

ValidationResult = Union[validators.ValidationFailure, bool]

GenericDict = Dict[K, V]

Configuration = Union[DictConfig, Dict[str, Any], argparse.Namespace]


def dir_path(path):
    """dir_path Type to use when parsing a path in argparse arguments


    Args:
        path (str): User provided path

    Raises:
        argparse.ArgumentTypeError: Path does not exists, so argparse fails

    Returns:
        str: User provided path

    Examples:
        >>> from slp.util.types import dir_path
        >>> import argparse
        >>> parser = argparse.ArgumentParser("My cool model")
        >>> parser.add_argument("--config", type=dir_path)
        >>> parser.parse_args(args=["--config", "my_random_config_that_does_not_exist.yaml"])
        Traceback (most recent call last):
        argparse.ArgumentTypeError: User provided path 'my_random_config_that_does_not_exist.yaml' does not exist

    """
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"User provided path '{path}' does not exist")



import errno
import itertools
import os
from collections import Counter
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import numpy as np
import slp.util.system as system
import slp.util.types as types
from loguru import logger
from slp.config.nlp import SPECIAL_TOKENS
from slp.data.transforms import HuggingFaceTokenizer, SpacyTokenizer, ToTokenIds
from tqdm import tqdm


def create_vocab(
    corpus: Union[List[str], List[List[str]]],
    vocab_size: int = -1,
    special_tokens: Optional[SPECIAL_TOKENS] = None,
) -> Dict[str, int]:
    """Create the vocabulary based on tokenized input corpus

    * Injects special tokens in the vocabulary
    * Calculates the occurence count for each token
    * Limits vocabulary to vocab_size most common tokens

    Args:
        corpus (Union[List[str], List[List[str]]]): The tokenized corpus as a list of sentences or a list of tokenized sentences
        vocab_size (int): [description]. Limit vocabulary to vocab_size most common tokens.
            Defaults to -1 which keeps all tokens.
        special_tokens Optional[SPECIAL_TOKENS]: Special tokens to include in the vocabulary. Defaults to None.

    Returns:
        Dict[str, int]: Dictionary of all accepted tokens and their corresponding occurence counts

    Examples:
        >>> create_vocab(["in", "a", "galaxy", "far", "far", "away"])
        {'far': 2, 'away': 1, 'galaxy': 1, 'a': 1, 'in': 1}
        >>> create_vocab(["in", "a", "galaxy", "far", "far", "away"], vocab_size=3)
        {'far': 2, 'a': 1, 'in': 1}
        >>> create_vocab(["in", "a", "galaxy", "far", "far", "away"], vocab_size=3, special_tokens=slp.config.nlp.SPECIAL_TOKENS)
        {'[PAD]': 0, '[MASK]': 0, '[UNK]': 0, '[BOS]': 0, '[EOS]': 0, '[CLS]': 0, '[SEP]': 0, 'far': 2, 'a': 1, 'in': 1}
    """
    if isinstance(corpus[0], list):
        corpus = list(itertools.chain.from_iterable(corpus))
    freq = Counter(corpus)
    if special_tokens is None:
        extra_tokens = []
    else:
        extra_tokens = special_tokens.to_list()
    if vocab_size < 0:
        vocab_size = len(freq)
    take = min(vocab_size, len(freq))
    logger.info(f"Keeping {vocab_size} most common tokens out of {len(freq)}")

    def take0(x: Tuple[Any, Any]) -> Any:
        """Take first tuple element"""
        return x[0]

    common_words = list(map(take0, freq.most_common(take)))
    common_words = list(set(common_words) - set(extra_tokens))
    words = extra_tokens + common_words
    if len(words) > vocab_size:
        words = words[: vocab_size + len(extra_tokens)]

    def token_freq(t):
        """Token frequeny"""
        return 0 if t in extra_tokens else freq[t]

    vocab = dict(zip(words, map(token_freq, words)))
    logger.info(f"Vocabulary created with {len(vocab)} tokens.")
    logger.info(f"The 10 most common tokens are:\n{freq.most_common(10)}")

    return vocab


class EmbeddingsLoader(object):
    def __init__(
        self,
        embeddings_file: str,
        dim: int,
        vocab: Optional[Dict[str, int]] = None,
        extra_tokens: Optional[SPECIAL_TOKENS] = None,
    ) -> None:
        """Load word embeddings in text format

        Args:
            embeddings_file (str): File where embeddings are stored (e.g. glove.6B.50d.txt)
            dim (int): Dimensionality of embeddings
            vocab (Optional[Dict[str, int]]): Load only embeddings in vocab. Defaults to None.
            extra_tokens (Optional[slp.config.nlp.SPECIAL_TOKENS]): Create random embeddings for these special tokens.
                Defaults to None.
        """
        self.embeddings_file = embeddings_file
        self.vocab = vocab
        self.cache_ = self._get_cache_name()
        self.dim_ = dim
        self.extra_tokens = extra_tokens

    def __repr__(self):
        """String representation of class"""
        return f"{self.__class__.__name__}({self.embeddings_file}, {self.dim_})"

    def in_accepted_vocab(self, word: str) -> bool:
        """Check if word exists in given vocabulary

        Args:
            word (str): word from embeddings file

        Returns:
            bool: Word exists
        """
        if self.vocab is None:
            return True
        else:
            return word in self.vocab

    def _get_cache_name(self) -> str:
        """Create a cache file name to avoid reloading the embeddings

        Cache name is something like glove.6B.50d.1000.p,
        where 1000 is the size of the vocab provided in __init__

        Returns:
            str: Cache file name
        """
        head, tail = os.path.split(self.embeddings_file)
        filename, ext = os.path.splitext(tail)
        if self.vocab is not None:
            cache_name = os.path.join(head, f"{filename}.{len(self.vocab)}.p")
        else:
            cache_name = os.path.join(head, f"{filename}.p")
        logger.info(f"Cache: {cache_name}")
        return cache_name

    def _dump_cache(self, data: types.Embeddings) -> None:
        """Save loaded embeddings to cache as a pickle

        Saves a tuple of (word2idx, idx2word, embeddings)

        Args:
            data (types.Embeddings): (word2idx, idx2word, embeddings) tuple
        """
        system.pickle_dump(data, self.cache_)

    def _load_cache(self) -> types.Embeddings:
        """Load Embeddings from cache

        Returns:
            types.Embeddings: (word2idx, idx2word, embeddings) tuple
        """
        return cast(types.Embeddings, system.pickle_load(self.cache_))

    def augment_embeddings(
        self,
        word2idx: Dict[str, int],
        idx2word: Dict[int, str],
        embeddings: List[np.ndarray],
        token: str,
        emb: Optional[np.ndarray] = None,
    ) -> Tuple[Dict[str, int], Dict[int, str], List[np.ndarray]]:
        """Create a random embedding for a special token and append it to the embeddings array

        Args:
            word2idx (Dict[str, int]): Current word2idx map
            idx2word (Dict[int, str]): Current idx2word map
            embeddings (List[np.ndarray]): Embeddings array as list of embeddings
            token (str): The special token (e.g. [PAD])
            emb (Optional[np.ndarray]): Optional value for the embedding to be appended.
                Defaults to None, where a random embedding is created.

        Returns:
            Tuple[Dict[str, int], Dict[int, str], List[np.ndarray]]: (word2idx, idx2word, embeddings) tuple
        """
        word2idx[token] = len(embeddings)
        idx2word[len(embeddings)] = token
        if emb is None:
            emb = np.random.uniform(low=-0.05, high=0.05, size=self.dim_)
        embeddings.append(emb)
        return word2idx, idx2word, embeddings

    @system.timethis(method=True)
    def load(self) -> types.Embeddings:
        """Read the word vectors from a text file

        * Read embeddings
        * Filter with given vocabulary
        * Augment with special tokens

        Returns:
            types.Embeddings: (word2idx, idx2word, embeddings) tuple
        """
        # in order to avoid this time consuming operation, cache the results
        try:
            cache = self._load_cache()
            logger.info("Loaded word embeddings from cache.")
            return cache
        except OSError:
            logger.warning(f"Didn't find embeddings cache file {self.embeddings_file}")
            logger.warning("Loading embeddings from file.")

        # create the necessary dictionaries and the word embeddings matrix
        if not os.path.exists(self.embeddings_file):
            logger.critical(f"{self.embeddings_file} not found!")
            raise OSError(errno.ENOENT, os.strerror(errno.ENOENT), self.embeddings_file)

        logger.info(f"Indexing file {self.embeddings_file} ...")

        # create the 2D array, which will be used for initializing
        # the Embedding layer of a NN.
        # We reserve the first row (idx=0), as the word embedding,
        # which will be used for zero padding (word with id = 0).
        if self.extra_tokens is not None:
            word2idx, idx2word, embeddings = self.augment_embeddings(
                {},
                {},
                [],
                self.extra_tokens.PAD.value,  # type: ignore
                emb=np.zeros(self.dim_),
            )

            for token in self.extra_tokens:  # type: ignore
                logger.debug(f"Adding token {token.value} to embeddings matrix")
                if token == self.extra_tokens.PAD:
                    continue
                word2idx, idx2word, embeddings = self.augment_embeddings(
                    word2idx, idx2word, embeddings, token.value
                )
        else:
            word2idx, idx2word, embeddings = self.augment_embeddings(
                {}, {}, [], "[PAD]", emb=np.zeros(self.dim_)
            )
        # read file, line by line
        with open(self.embeddings_file, "r") as f:
            num_lines = sum(1 for line in f)

        with open(self.embeddings_file, "r") as f:
            index = len(embeddings)

            for line in tqdm(
                f, total=num_lines, desc="Loading word embeddings...", leave=False
            ):
                # skip the first row if it is a header
                if len(line.split()) < self.dim_:
                    continue

                values = line.rstrip().split(" ")
                word = values[0]

                if word in word2idx:
                    continue

                if not self.in_accepted_vocab(word):
                    continue

                vector = np.asarray(values[1:], dtype=np.float32)
                idx2word[index] = word
                word2idx[word] = index
                embeddings.append(vector)
                index += 1

        logger.info(f"Loaded {len(embeddings)} word vectors.")
        embeddings_out = np.array(embeddings, dtype="float32")

        # write the data to a cache file
        self._dump_cache((word2idx, idx2word, embeddings_out))
        return word2idx, idx2word, embeddings_out


class WordCorpus(object):
    def __init__(
        self,
        corpus: List[str],
        limit_vocab_size: int = 30000,
        word2idx: Optional[Dict[str, int]] = None,
        idx2word: Optional[Dict[int, str]] = None,
        embeddings: Optional[np.ndarray] = None,
        embeddings_file: Optional[str] = None,
        embeddings_dim: int = 300,
        lower: bool = True,
        special_tokens: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
        prepend_bos: bool = False,
        append_eos: bool = False,
        lang: str = "en_core_web_md",
        max_len: int = -1,
        **kwargs,
    ):
        """Load corpus embeddings, tokenize in words using spacy and convert to ids

        This class handles the handling of a raw corpus. It handles:

        * Tokenization into words (spacy)
        * Loading of pretrained word embedding
        * Calculation of word frequencies / corpus statistics
        * Conversion to token ids

        You can pass either:

        * Pass an embeddings file to load pretrained embeddings and create the word2idx mapping
        * Pass already loaded embeddings array and word2idx. This is useful for the dev / test splits
          where we want to pass the train split embeddings / word2idx.

        Args:
            corpus (List[List[str]]): Corpus as a list of sentences
            limit_vocab_size (int): Upper bound for number of most frequent tokens to keep. Defaults to 30000.
            word2idx (Optional[Dict[str, int]]): Mapping of word to indices. Defaults to None.
            idx2word (Optional[Dict[int, str]]): Mapping of indices to words. Defaults to None.
            embeddings (Optional[np.ndarray]): Embeddings array. Defaults to None.
            embeddings_file (Optional[str]): Embeddings file to read. Defaults to None.
            embeddings_dim (int): Dimension of embeddings. Defaults to 300.
            lower (bool): Convert strings to lower case. Defaults to True.
            special_tokens (Optional[SPECIAL_TOKENS]): Special tokens to include in the vocabulary.
                 Defaults to slp.config.nlp.SPECIAL_TOKENS.
            prepend_bos (bool): Prepend Beginning of Sequence token for seq2seq tasks. Defaults to False.
            append_eos (bool): Append End of Sequence token for seq2seq tasks. Defaults to False.
            lang (str): Spacy language, e.g. el_core_web_sm, en_core_web_sm etc. Defaults to "en_core_web_md".
            max_len (int): Crop sequences above this length. Defaults to -1 where sequences are left unaltered.
        """
        # FIXME: Extract super class to avoid repetition
        self.corpus_ = corpus
        self.max_len = max_len
        self.tokenizer = SpacyTokenizer(
            lower=lower,
            prepend_bos=prepend_bos,
            append_eos=append_eos,
            specials=special_tokens,
            lang=lang,
        )

        logger.info(f"Tokenizing corpus using spacy {lang}")

        self.tokenized_corpus_ = [
            self.tokenizer(s)
            for s in tqdm(self.corpus_, desc="Tokenizing corpus...", leave=False)
        ]

        self.vocab_ = create_vocab(
            self.tokenized_corpus_,
            vocab_size=limit_vocab_size if word2idx is None else -1,
            special_tokens=special_tokens,
        )

        self.word2idx_, self.idx2word_, self.embeddings_ = None, None, None
        # self.corpus_indices_ = self.tokenized_corpus_

        if word2idx is not None:
            logger.info("Word2idx was already provided. Going to used it.")

        if embeddings_file is not None and word2idx is None:
            logger.info(
                f"Going to load {len(self.vocab_)} embeddings from {embeddings_file}"
            )
            loader = EmbeddingsLoader(
                embeddings_file,
                embeddings_dim,
                vocab=self.vocab_,
                extra_tokens=special_tokens,
            )
            word2idx, idx2word, embeddings = loader.load()

        if embeddings is not None:
            self.embeddings_ = embeddings

        if idx2word is not None:
            self.idx2word_ = idx2word

        if word2idx is not None:
            self.word2idx_ = word2idx

            logger.info("Converting tokens to ids using word2idx.")
            self.to_token_ids = ToTokenIds(
                self.word2idx_,
                specials=SPECIAL_TOKENS,  # type: ignore
            )

            self.corpus_indices_ = [
                self.to_token_ids(s)
                for s in tqdm(
                    self.tokenized_corpus_,
                    desc="Converting tokens to token ids...",
                    leave=False,
                )
            ]

            logger.info("Filtering corpus vocabulary.")

            updated_vocab = {}
            for k, v in self.vocab_.items():
                if k in self.word2idx_:
                    updated_vocab[k] = v

            logger.info(
                f"Out of {len(self.vocab_)} tokens {len(self.vocab_) - len(updated_vocab)} were not found in the pretrained embeddings."
            )

            self.vocab_ = updated_vocab

    @property
    def vocab_size(cls) -> int:
        """Retrieve vocabulary size for corpus

        Returns:
            int: vocabulary size
        """
        return (
            cls.embeddings.shape[0] if cls.embeddings is not None else len(cls.vocab_)
        )

    @property
    def frequencies(cls) -> Dict[str, int]:
        """Retrieve word occurence counts

        Returns:
            Dict[str, int]: word occurence counts
        """
        return cls.vocab_

    @property
    def vocab(cls) -> Set[str]:
        """Retrieve set of words in vocabulary

        Returns:
            Set[str]: set of words in vocabulary
        """
        return set(cls.vocab_.keys())

    @property
    def embeddings(cls) -> np.ndarray:
        """Retrieve embeddings array

        Returns:
            np.ndarray: Array of pretrained word embeddings
        """
        return cast(np.ndarray, cls.embeddings_)

    @property
    def word2idx(cls) -> Dict[str, int]:
        """Retrieve word2idx mapping

        Returns:
            Dict[str, int]: word2idx mapping
        """
        return cast(Dict[str, int], cls.word2idx_)

    @property
    def idx2word(cls) -> Dict[int, str]:
        """Retrieve idx2word mapping

        Returns:
            Dict[str, int]: idx2word mapping
        """
        return cast(Dict[int, str], cls.idx2word_)

    @property
    def tokenized(cls) -> List[List[str]]:
        """Retrieve tokenized corpus

        Returns:
            List[List[str]]: Tokenized corpus
        """
        return cls.tokenized_corpus_

    @property
    def indices(cls) -> List[List[int]]:
        """Retrieve corpus as token indices

        Returns:
            List[List[int]]: Token indices for corpus
        """
        return cls.corpus_indices_

    @property
    def raw(cls) -> List[str]:
        """Retrieve raw corpus

        Returns:
            List[str]: Raw Corpus
        """
        return cls.corpus_

    def __len__(self) -> int:
        """Number of samples in corpus

        Returns:
            int: Corpus length
        """
        return len(self.corpus_indices_)

    def __getitem__(self, idx) -> List[int]:
        """Get ith element in corpus as token indices

        Args:
            idx (List[int]): index in corpus

        Returns:
            List[int]: List of token indices for sentence
        """
        indices = self.corpus_indices_[idx]
        out: List[int] = (
            self.corpus_indices_[idx]
            if self.max_len <= 0
            else self.corpus_indices_[idx][: self.max_len]
        )
        return out


class HfCorpus(object):
    def __init__(
        self,
        corpus: List[str],
        lower: bool = True,
        tokenizer_model: str = "bert-base-uncased",
        add_special_tokens: bool = True,
        special_tokens: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
        max_len: int = -1,
        **kwargs,
    ):
        """Process a corpus using hugging face tokenizers

        Select one of hugging face tokenizers and process corpus

        Args:
            corpus (List[str]): List of sentences
            lower (bool): Convert strings to lower case. Defaults to True.
            tokenizer_model (str): Hugging face model to use. Defaults to "bert-base-uncased".
            add_special_tokens (bool): Add special tokens in sentence during tokenization. Defaults to True.
            special_tokens (Optional[SPECIAL_TOKENS]): Special tokens to include in the vocabulary.
                 Defaults to slp.config.nlp.SPECIAL_TOKENS.
            max_len (int): Crop sequences above this length. Defaults to -1 where sequences are left unaltered.
        """
        self.corpus_ = corpus
        self.max_len = max_len

        logger.info(
            f"Tokenizing corpus using hugging face tokenizer from {tokenizer_model}"
        )

        self.tokenizer = HuggingFaceTokenizer(
            lower=lower, model=tokenizer_model, add_special_tokens=add_special_tokens
        )

        self.corpus_indices_ = [
            self.tokenizer(s)
            for s in tqdm(
                self.corpus_, desc="Converting tokens to indices...", leave=False
            )
        ]

        self.tokenized_corpus_ = [
            self.tokenizer.detokenize(s)
            for s in tqdm(
                self.corpus_indices_,
                desc="Mapping indices to tokens...",
                leave=False,
            )
        ]

        self.vocab_ = create_vocab(
            self.tokenized_corpus_,
            vocab_size=-1,
            special_tokens=special_tokens,
        )

    @property
    def vocab_size(cls) -> int:
        """Retrieve vocabulary size

        Returns:
            int: Vocabulary size
        """
        sz: int = cls.tokenizer.vocab_size
        return sz

    @property
    def frequencies(cls) -> Dict[str, int]:
        """Retrieve wordpieces occurence counts

        Returns:
            Dict[str, int]: wordpieces occurence counts
        """
        return cls.vocab_

    @property
    def vocab(cls) -> Set[str]:
        """Retrieve set of words in vocabulary

        Returns:
            Set[str]: set of words in vocabulary
        """
        return set(cls.vocab_.keys())

    @property
    def embeddings(cls) -> None:
        """Unused. Defined for compatibility"""
        return None

    @property
    def word2idx(cls) -> None:
        """Unused. Defined for compatibility"""
        return None

    @property
    def idx2word(cls) -> None:
        """Unused. Defined for compatibility"""
        return None

    @property
    def tokenized(cls) -> List[List[str]]:
        """Retrieve tokenized corpus

        Returns:
            List[List[str]]: tokenized corpus
        """
        return cls.tokenized_corpus_

    @property
    def indices(cls) -> List[List[int]]:
        """Retrieve corpus as token indices

        Returns:
            List[List[int]]: Token indices for corpus
        """
        return cls.corpus_indices_

    @property
    def raw(cls) -> List[str]:
        """Retrieve raw corpus

        Returns:
            List[str]: Raw Corpus
        """
        return cls.corpus_

    def __len__(self) -> int:
        """Number of samples in corpus

        Returns:
            int: Corpus length
        """
        return len(self.corpus_indices_)

    def __getitem__(self, idx):
        """Get ith element in corpus as token indices

        Args:
            idx (List[int]): index in corpus

        Returns:
            List[int]: List of token indices for sentence
        """
        return (
            self.corpus_indices_[idx]
            if self.max_len <= 0
            else self.corpus_indices_[idx][: self.max_len]
        )


class TokenizedCorpus(object):
    def __init__(
        self,
        corpus: Union[List[str], List[List[str]]],
        word2idx: Dict[str, int] = None,
        special_tokens: Optional[SPECIAL_TOKENS] = SPECIAL_TOKENS,  # type: ignore
        max_len: int = -1,
        **kwargs,
    ):
        """Wrap a corpus that's already tokenized

        Args:
            corpus (Union[List[str], List[List[str]]]): List of tokens or List of lists of tokens
            word2idx (Dict[str, int], optional): Token to index mapping. Defaults to None.
            special_tokens (Optional[SPECIAL_TOKENS], optional): Special Tokens. Defaults to SPECIAL_TOKENS.
        """
        self.corpus_ = corpus
        self.tokenized_corpus_ = corpus
        self.max_len = max_len

        self.vocab_ = create_vocab(
            self.tokenized_corpus_,
            vocab_size=-1,
            special_tokens=special_tokens,
        )

        if word2idx is not None:
            logger.info("Converting tokens to ids using word2idx.")
            self.word2idx_ = word2idx
        else:
            logger.info(
                "No word2idx provided. Will convert tokens to ids using an iterative counter."
            )
            self.word2idx_ = dict(zip(self.vocab_.keys(), itertools.count()))

        self.idx2word_ = {v: k for k, v in self.word2idx_.items()}

        self.to_token_ids = ToTokenIds(
            self.word2idx_,
            specials=SPECIAL_TOKENS,  # type: ignore
        )
        if isinstance(self.tokenized_corpus_[0], list):
            self.corpus_indices_ = [
                self.to_token_ids(s)
                for s in tqdm(
                    self.tokenized_corpus_,
                    desc="Converting tokens to token ids...",
                    leave=False,
                )
            ]
        else:
            self.corpus_indices_ = self.to_token_ids(self.tokenized_corpus_)  # type: ignore

    @property
    def vocab_size(cls) -> int:
        """Retrieve vocabulary size

        Returns:
            int: Vocabulary size
        """
        return len(cls.vocab_)

    @property
    def frequencies(cls) -> Dict[str, int]:
        """Retrieve wordpieces occurence counts

        Returns:
            Dict[str, int]: wordpieces occurence counts
        """
        return cls.vocab_

    @property
    def vocab(cls) -> Set[str]:
        """Retrieve set of words in vocabulary

        Returns:
            Set[str]: set of words in vocabulary
        """
        return set(cls.vocab_.keys())

    @property
    def embeddings(cls) -> None:
        """Unused. Kept for compatibility"""
        return None

    @property
    def word2idx(cls) -> Dict[str, int]:
        """Retrieve word2idx mapping

        Returns:
            Dict[str, int]: word2idx mapping
        """
        return cls.word2idx_

    @property
    def idx2word(cls) -> Dict[int, str]:
        """Retrieve idx2word mapping

        Returns:
            Dict[str, int]: idx2word mapping
        """
        return cls.idx2word_

    @property
    def tokenized(cls) -> Union[List[str], List[List[str]]]:
        """Retrieve tokenized corpus

        Returns:
            List[List[str]]: Tokenized corpus
        """
        return cls.tokenized_corpus_

    @property
    def indices(cls) -> Union[List[int], List[List[int]]]:
        """Retrieve corpus as token indices

        Returns:
            List[List[int]]: Token indices for corpus
        """
        return cls.corpus_indices_

    @property
    def raw(cls) -> Union[List[str], List[List[str]]]:
        """Retrieve raw corpus

        Returns:
            List[str]: Raw Corpus
        """
        return cls.corpus_

    def __len__(self) -> int:
        """Number of samples in corpus

        Returns:
            int: Corpus length
        """
        return len(self.corpus_indices_)

    def __getitem__(self, idx) -> List[int]:
        """Get ith element in corpus as token indices

        Args:
            idx (List[int]): index in corpus

        Returns:
            List[int]: List of token indices for sentence
        """
        out: List[int] = (
            self.corpus_indices_[idx]
            if self.max_len <= 0
            else self.corpus_indices_[idx][: self.max_len]
        )
        return out


if __name__ == "__main__":
    corpus = [
        "the big",
        "brown fox",
        "jumps over",
        "the lazy dog",
        "supercalifragilisticexpialidocious",
    ]

    word_corpus = WordCorpus(
        corpus,
        embeddings_file="./cache/glove.6B.50d.txt",
        embeddings_dim=50,
        lower=True,
        prepend_bos=True,
        append_eos=True,
    )

    hugging_face_corpus = HfCorpus(corpus)



import math

import torch.nn as nn
import torch.optim as optim
from slp.config.nlp import SPECIAL_TOKENS
from slp.data import Seq2SeqCollator
from slp.modules.embed import PositionalEncoding
from slp.modules.transformer import Encoder as TransformerEncoder
from slp.plbind import (
    PLDataModuleFromCorpus,
    TransformerPLModule,
    make_trainer,
    watch_model,
)
from slp.util.log import configure_logging
from torchnlp.datasets import wikitext_2_dataset  # type: ignore
from torchnlp.samplers import BPTTBatchSampler


class TransformerLM(nn.Module):
    def __init__(
        self,
        vocab_size=30000,
        num_layers=2,
        hidden_size=200,
        num_heads=2,
        inner_size=256,
        dropout=0.2,
        tie_weights=True,
    ):
        super(TransformerLM, self).__init__()
        self.pos_encoder = PositionalEncoding(hidden_size, max_len=5000)
        self.transformer_encoder = TransformerEncoder(
            num_layers=num_layers,
            hidden_size=hidden_size,
            num_heads=num_heads,
            inner_size=inner_size,
            dropout=dropout,
        )
        self.hidden_size = hidden_size
        self.encoder = nn.Embedding(vocab_size, hidden_size)
        self.decoder = nn.Linear(hidden_size, vocab_size)
        self.tie_weights = tie_weights

        if tie_weights:
            self.decoder.weight = self.encoder.weight

        self.init_weights()

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.encoder.weight, -initrange, initrange)

        if not self.tie_weights:
            nn.init.zeros_(self.decoder.weight)
            nn.init.uniform_(self.decoder.weight, -initrange, initrange)

    def forward(self, src, targets, source_mask=None, target_mask=None):
        src = self.encoder(src) * math.sqrt(self.hidden_size)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, attention_mask=target_mask)
        output = self.decoder(output)

        return output


collate_fn = Seq2SeqCollator(device="cpu")


if __name__ == "__main__":
    EXPERIMENT_NAME = "transformer-wikitext2"
    configure_logging(f"logs/{EXPERIMENT_NAME}")

    bptt = 35  # TODO: argparse this
    vocab_size = -1
    lr = 1e-4

    train, dev, test = wikitext_2_dataset(
        directory="data/",
        train=True,
        dev=True,
        test=True,
        extracted_name="wikitext-2",
        url="https://s3.amazonaws.com/research.metamind.io/wikitext/wikitext-2-v1.zip",  # noqa: E501
        unknown_token=SPECIAL_TOKENS.UNK.value,
        eos_token=SPECIAL_TOKENS.EOS.value,
    )

    ldm = PLDataModuleFromCorpus(
        train,
        val=dev,
        test=test,
        drop_last=True,
        max_len=-1,
        batch_sampler_train=BPTTBatchSampler(train, bptt, 20, True),
        batch_sampler_val=BPTTBatchSampler(dev, bptt, 10, True),
        batch_sampler_test=BPTTBatchSampler(test, bptt, 10, True),
        pin_memory=True,
        num_workers=0,
        language_model=True,
        tokenizer="tokenized",
        collate_fn=collate_fn,
    )

    model = TransformerLM(
        vocab_size=ldm.vocab_size,
        num_layers=2,
        hidden_size=200,
        num_heads=2,
        inner_size=256,
        dropout=0.2,
        tie_weights=True,
    )

    optimizer = optim.Adam([p for p in model.parameters() if p.requires_grad], lr=lr)
    criterion = nn.CrossEntropyLoss()

    lm = TransformerPLModule(
        model,
        optimizer,
        criterion,
        calculate_perplexity=True,
    )

    trainer = make_trainer(
        EXPERIMENT_NAME, max_epochs=100, gpus=1, gradient_clip_val=0.25
    )
    watch_model(trainer, model)

    trainer.fit(lm, datamodule=ldm)

    trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())



import argparse
import sys
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

from omegaconf import DictConfig, OmegaConf


def _nest(d: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """_nest Recursive function to nest a dictionary on keys with . (dots)

    Parse documentation into a hierarchical dict. Keys should be separated by dots (e.g. "model.hidden") to go down into the hierarchy

    Args:
        d (Dict[str, Any]): dictionary containing flat config values

    Returns:
        Dict[str, Any]: Hierarchical config dictionary

    Examples:
        >>> _nest({{"model.hidden": 20, "optimizer.lr": 1e-3}})
        {"model": {"hidden": 20}, "optimizer": {"lr": 1e-3}}
    """
    nested: Dict[str, Any] = defaultdict(dict)

    for key, val in d.items():
        if "." in key:
            splitkeys = key.split(".")
            inner = _nest({".".join(splitkeys[1:]): val})

            if inner is not None:
                nested[splitkeys[0]].update(inner)
        else:
            if val is not None:
                nested[key] = val

    return dict(nested) if nested else None


class OmegaConfExtended(OmegaConf):
    """OmegaConfExtended Extended OmegaConf class, to include argparse style CLI arguments

    Unfortunately the original authors are not interested into providing integration with argparse
    (https://github.com/omry/omegaconf/issues/569), so we have to get by with this extension
    """

    @staticmethod
    def from_argparse(
        parser: argparse.ArgumentParser, args: Optional[List[str]] = None
    ) -> Tuple[DictConfig, DictConfig]:
        """from_argparse Static method to convert argparse arguments into OmegaConf DictConfig objects

        We parse the command line arguments and separate the user provided values and the default values.
        This is useful for merging with a config file.

        Args:
            parser (argparse.ArgumentParser): Parser for argparse arguments
            args (Optional[List[str]]): Optional input sys.argv style args. Useful for testing.
                Use this only for testing. By default it uses sys.argv[1:]
        Returns:
            Tuple[omegaconf.DictConfig, omegaconf.DictConfig]: (user provided cli args, default cli args) as a tuple of omegaconf.DictConfigs

        Examples:
            >>> import argparse
            >>> from slp.config.omegaconf import OmegaConfExtended
            >>> parser = argparse.ArgumentParser("My cool model")
            >>> parser.add_argument("--hidden", dest="model.hidden", type=int, default=20)
            >>> user_provided_args, default_args = OmegaConfExtended.from_argparse(parser, args=["--hidden", "100"])
            >>> user_provided_args
            {'model': {'hidden': 100}}
            >>> default_args
            {}
            >>> user_provided_args, default_args = OmegaConfExtended.from_argparse(parser)
            >>> user_provided_args
            {}
            >>> default_args
            {'model': {'hidden': 20}}
        """
        dest_to_arg = {v.dest: k for k, v in parser._option_string_actions.items()}

        all_args = vars(parser.parse_args(args=args))
        provided_args = {}
        default_args = {}

        for k, v in all_args.items():
            if dest_to_arg[k] in sys.argv:
                provided_args[k] = v
            else:
                default_args[k] = v

        provided = OmegaConf.create(_nest(provided_args))
        defaults = OmegaConf.create(_nest(default_args))

        return provided, defaults



# python examples/mnist.py --bsz 128 --bsz-eval 256

import sys
from argparse import ArgumentParser

import pytorch_lightning as pl
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from loguru import logger
from slp.config.config_parser import make_cli_parser, parse_config
from slp.plbind import (
    FromLogits,
    PLDataModuleFromDatasets,
    PLModule,
    make_trainer,
    watch_model,
)
from slp.util.log import configure_logging
from torchvision.datasets import MNIST  # type: ignore
from torchvision.transforms import Compose, Normalize, ToTensor  # type: ignore

# Could be read from yaml with OmegaConf.from_yaml
CONFIG = {
    "model": {"intermediate_hidden": 100},
    "optimizer": "Adam",
    "optim": {"lr": 1e-3},
}


class Net(nn.Module):
    def __init__(self, intermediate_hidden=50):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, intermediate_hidden)
        self.fc2 = nn.Linear(intermediate_hidden, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return x


def get_data():
    data_transform = Compose([ToTensor(), Normalize((0.1307,), (0.3081,))])
    train = MNIST(download=True, root=".", transform=data_transform, train=True)

    val = MNIST(download=False, root=".", transform=data_transform, train=False)
    return train, val


def get_parser():
    parser = ArgumentParser("MNIST classification example")
    parser.add_argument(
        "--hidden",
        dest="model.intermediate_hidden",
        type=int,
        default=12,
        help="Intermediate hidden layers for linear module",
    )
    return parser


if __name__ == "__main__":
    # SETUP ##################################################
    parser = get_parser()
    parser = make_cli_parser(parser, PLDataModuleFromDatasets)

    config = parse_config(parser, parser.parse_args().config)

    if config.trainer.experiment_name == "experiment":
        config.trainer.experiment_name = "mnist-classification"

    configure_logging(f"logs/{config.trainer.experiment_name}")

    if config.seed is not None:
        logger.info("Seeding everything with seed={seed}")
        pl.utilities.seed.seed_everything(seed=config.seed)

    # Get data and make datamodule ##########################
    train, test = get_data()

    ldm = PLDataModuleFromDatasets(train, test=test, seed=config.seed, **config.data)

    # Create model, optimizer, criterion, scheduler ###########
    model = Net(**config.model)

    optimizer = getattr(optim, config.optimizer)(model.parameters(), **config.optim)
    criterion = nn.CrossEntropyLoss()

    lr_scheduler = None
    if config.lr_scheduler:
        lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, **config.lr_schedule
        )

    # Wrap in PLModule, & configure metrics ####################
    lm = PLModule(
        model,
        optimizer,
        criterion,
        lr_scheduler=lr_scheduler,
        metrics={"acc": FromLogits(pl.metrics.classification.Accuracy())},
        hparams=config,
    )

    # Run debugging session or fit & test the model ############
    if config.debug:
        logger.info("Running in debug mode: Fast run on 5 batches")
        trainer = make_trainer(fast_dev_run=5)
        trainer.fit(lm, datamodule=ldm)

        logger.info("Running in debug mode: Overfitting 5 batches")
        trainer = make_trainer(overfit_batches=5)
        trainer.fit(lm, datamodule=ldm)

    else:
        trainer = make_trainer(**config.trainer)
        watch_model(trainer, model)

        trainer.fit(lm, datamodule=ldm)

        trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())

        logger.info("Run finished. Uploading files to wandb...")



import functools
import os
import pickle
import shutil
import socket
import subprocess
import sys
import time
import urllib
import urllib.request
from datetime import datetime
from typing import Any, Callable, Optional, Tuple, cast

import validators
from loguru import logger
from slp.util import types

try:
    import ujson as json
except ImportError:
    import json  # type: ignore


def has_internet_connection(timeout: int = 3) -> bool:
    """has_internet_connection Check if you are connected to the internet

    Check if internet connection exists by pinging Google DNS server

    Host: 8.8.8.8 (google-public-dns-a.google.com)
    OpenPort: 53/tcp
    Service: domain (DNS/TCP)

    Args:
        timeout (int): Seconds to wait before giving up

    Returns:
        bool: True if connection is established, False if we are not connected to the internet
    """
    host, port = "8.8.8.8", 53
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error as ex:
        print(ex)
        return False


def date_fname() -> str:
    """date_fname Generate a filename based on datetime.now().

    If multiple calls are made within the same second, the filename will not be unique.
    We could add miliseconds etc. in the fname but that would hinder readability.
    For practical purposes e.g. unique logs between different experiments this should be enough.
    Either way if we need a truly unique descriptor, there is the uuid module.

    Returns:
        str: A filename, e.g. 20210228-211832
    """
    return datetime.now().strftime("%Y%m%d-%H%M%S")


def print_separator(
    symbol: str = "*", n: int = 10, print_fn: Callable[[str], None] = print
):
    """print_separator Print a repeated symbol as a separator

    *********************************************************

    Args:
        symbol (str): Symbol to print
        n (int): Number of times to print the symbol
        print_fn (Callable[[str], None]): Print function to use, e.g. print or logger.info

    Examples:
        >>> print_separator(symbol="-", n=2)
        --
    """
    print_fn(symbol * n)


def is_url(inp: Optional[str]) -> types.ValidationResult:
    """is_url Check if the provided string is a URL

    Args:
        inp (Optional[str]): A potential link or None

    Returns:
        types.ValidationResult: True if a valid url is provided, False if the string is not a url

    Examples:
        >>> is_url("Hello World")
        ValidationFailure(func=url, args={'value': 'Hello World', 'public': False})
        >>> is_url("http://google.com")
        True
    """
    if not inp:
        return False
    return validators.url(inp)


def is_file(inp: Optional[str]) -> types.ValidationResult:
    """is_file Check if the provided string is valid file in the system path

    Args:
        inp (Optional[str]): A potential file or None

    Returns:
        types.ValidationResult: True if a valid file is provided, False if the string is not a url

    Examples:
        >>> is_file("/bin/bash")
        True
        >>> is_file("/supercalifragilisticexpialidocious")  # This does not exist. I hope...
        False
    """
    if not inp:
        return False
    return os.path.isfile(inp)


def is_subpath(child: str, parent: str) -> bool:
    """is_subpath Check if child path is a subpath of parent

    Args:
        child (str): Child path
        parent (str): parent path

    Returns:
        bool: True if child is a subpath of parent, false if not

    Examples:
        >>> is_subpath("/usr/bin/Xorg", "/usr")
        True
    """
    parent = os.path.abspath(parent)
    child = os.path.abspath(child)
    return cast(
        bool, os.path.commonpath([parent]) == os.path.commonpath([parent, child])
    )


def safe_mkdirs(path: str) -> None:
    """safe_mkdirs Makes recursively all the directories in input path

    Utility function similar to mkdir -p. Makes directories recursively, if given path does not exist

    Args:
        path (str): Path to mkdir -p

    Examples:
        >>> safe_mkdirs("super/cali/fragi/listic/expi/ali/docious")
    """
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except Exception as e:
            logger.warning(e)
            raise IOError((f"Failed to create recursive directories: {path}"))


def timethis(method=False) -> Callable:
    """timethis Decorator to measure the time it takes for a function to complete

    Examples:
        >>> @slp.util.system.timethis
        >>> def time_consuming_function(...): ...
    """

    def timethis_inner(func: Callable) -> Callable:
        """Inner function for decorator closure"""

        @functools.wraps(func)
        def timed(*args: types.T, **kwargs: types.T):
            """Inner function for decorator closure"""

            ts = time.time()
            result = func(*args, **kwargs)
            te = time.time()
            elapsed = f"{te - ts}"
            if method:

                logger.info(
                    "BENCHMARK: {cls}.{f}(*{a}, **{kw}) took: {t} sec".format(
                        f=func.__name__, cls=args[0], a=args[1:], kw=kwargs, t=elapsed
                    )
                )
            else:
                logger.info(
                    "BENCHMARK: {f}(*{a}, **{kw}) took: {t} sec".format(
                        f=func.__name__, a=args, kw=kwargs, t=elapsed
                    )
                )
            return result

        return cast(Callable, timed)

    return timethis_inner


def suppress_print(func: Callable) -> Callable:
    """suppress_print Decorator to supress stdout of decorated function

    Examples:
        >>> @slp.util.system.timethis
        >>> def very_verbose_function(...): ...
    """

    def func_wrapper(*args: types.T, **kwargs: types.T):
        """Inner function for decorator closure"""
        with open("/dev/null", "w") as sys.stdout:
            ret = func(*args, **kwargs)
        sys.stdout = sys.__stdout__
        return ret

    return cast(Callable, func_wrapper)


def run_cmd(command: str) -> Tuple[int, str]:
    """run_cmd Run given shell command

    Args:
        command (str): Shell command to run

    Returns:
        (int, str): Status code, stdout of shell command

    Examples:
        >>> run_cmd("ls /")
        (0, 'bin\nboot\ndev\netc\nhome\ninit\nlib\nlib32\nlib64\nlibx32\nlost+found\nmedia\nmnt\nopt\nproc\nroot\nrun\nsbin\nsnap\nsrv\nsys\ntmp\nusr\nvar\n')
    """
    command = f'{os.getenv("SHELL")} -c "{command}"'
    pipe = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )

    stdout = ""
    if pipe.stdout is not None:
        stdout = "".join(
            [line.decode("utf-8") for line in iter(pipe.stdout.readline, b"")]
        )
        pipe.stdout.close()
    returncode = pipe.wait()
    return returncode, stdout


def run_cmd_silent(command: str) -> Tuple[int, str]:
    """run_cmd_silent Run command without printing to console

    Args:
        command (str): Shell command to run

    Returns:
        (int, str): Status code, stdout of shell command

    Examples:
        >>> run_cmd("ls /")
        (0, 'bin\nboot\ndev\netc\nhome\ninit\nlib\nlib32\nlib64\nlibx32\nlost+found\nmedia\nmnt\nopt\nproc\nroot\nrun\nsbin\nsnap\nsrv\nsys\ntmp\nusr\nvar\n')
    """
    return cast(Tuple[int, str], suppress_print(run_cmd)(command))


def download_url(url: str, dest_path: str) -> str:
    """download_url Download a file to a destination path given a URL

    Args:
        url (str): A url pointing to the file we want to download
        dest_path (str): The destination path to write the file

    Returns:
        (str): The filename where the downloaded file is written
    """
    name = url.rsplit("/")[-1]
    dest = os.path.join(dest_path, name)
    safe_mkdirs(dest_path)
    response = urllib.request.urlopen(url)
    with open(dest, "wb") as fd:
        shutil.copyfileobj(response, fd)
    return dest


def write_wav(byte_str: str, wav_file: str) -> None:
    """write_wav Write a hex string into a wav file

    Args:
        byte_str (str): The hex string containing the audio data
        wav_file (str): The output wav file
    """
    with open(wav_file, "w") as fd:
        fd.write(byte_str)


def read_wav(wav_sample: str) -> str:
    """read_wav Reads a wav clip into a string and returns the hex string.

    Args:
        wav_sample (str): Path to wav file

    Returns:
        A hex string with the audio information.
    """
    with open(wav_sample, "r") as wav_fd:
        clip = wav_fd.read()
    return clip


def pickle_load(fname: str) -> Any:
    """pickle_load Load data from pickle file

    Args:
        fname (str): file name of pickle file

    Returns:
        Any: Loaded data
    """
    with open(fname, "rb") as fd:
        data = pickle.load(fd)
    return data


def pickle_dump(data: Any, fname: str) -> None:
    """pickle_dump Save data to pickle file

    Args:
        data (Any): Data to save
        fname (str): Output pickle file
    """
    with open(fname, "wb") as fd:
        pickle.dump(data, fd)


def json_load(fname: str) -> types.GenericDict:
    """json_load Load dict from a json file

    Args:
        fname (str): Json file to load

    Returns:
        types.GenericDict: Dict of loaded data
    """
    with open(fname, "r") as fd:
        data = json.load(fd)
    return cast(types.GenericDict, data)


def json_dump(data: types.GenericDict, fname: str) -> None:
    """json_dump Save dict to a json file

    Args:
        data (types.GenericDict): Dict to save
        fname (str): Output json file
    """
    with open(fname, "w") as fd:
        json.dump(data, fd)



import os
from typing import Optional

import pytorch_lightning as pl
import torch
import torch.nn.functional as F
import wandb
from loguru import logger
from pytorch_lightning.utilities import rank_zero_only
from slp.util.system import print_separator


class EarlyStoppingWithLogs(pl.callbacks.EarlyStopping):
    def _run_early_stopping_check(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ):
        """Check if we should early stop on the early_stop_metric and print logs

        Args:
            trainer (pl.Trainer): Trainer
            pl_module (pl.LightningModule): Module used for training
        """
        super(EarlyStoppingWithLogs, self)._run_early_stopping_check(trainer, pl_module)
        logger.info(f"Epoch {trainer.current_epoch + 1} Early Stopping")
        print_separator(symbol="-", n=50, print_fn=logger.info)
        logger.info("{:<15} {:<15}".format("best score", self.best_score))
        logger.info(
            "{:<15} {:<15}".format("patience left", self.patience - self.wait_count)
        )

        if trainer.should_stop:  # type: ignore
            logger.info("Stopping due to early stopping")
        print_separator(symbol="#", n=50, print_fn=logger.info)


class FromLogits(pl.metrics.Metric):
    def __init__(self, metric: pl.metrics.Metric):
        """Wrap pytorch lighting metric to accept logits input

        Args:
            metric (pl.metrics.Metric): The metric to wrap, e.g. pl.metrics.Accuracy
        """
        super(FromLogits, self).__init__(
            compute_on_step=metric.compute_on_step,
            dist_sync_on_step=metric.dist_sync_on_step,
            process_group=metric.process_group,
            dist_sync_fn=metric.dist_sync_fn,
        )
        self.metric = metric

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:  # type: ignore
        """Update underlying metric

        Calculate softmax under the hood and pass probs to the underlying metric

        Args:
            preds (torch.Tensor): [B, *, num_classes] Logits
            target (torch.Tensor): [B, *] Ground truths
        """
        preds = F.softmax(preds, dim=-1)
        self.metric.update(preds, target)  # type: ignore

    def compute(self) -> torch.Tensor:
        """Compute metric

        Returns:
            torch.Tensor: metric value
        """
        return self.metric.compute()  # type: ignore


class FixedWandbLogger(pl.loggers.WandbLogger):
    def __init__(
        self,
        name: Optional[str] = None,
        save_dir: Optional[str] = None,
        offline: Optional[bool] = False,
        id: Optional[str] = None,
        anonymous: Optional[bool] = False,
        version: Optional[str] = None,
        project: Optional[str] = None,
        log_model: Optional[bool] = False,
        experiment: wandb.sdk.wandb_run.Run = None,
        prefix: Optional[str] = "",
        sync_step: Optional[bool] = True,
        checkpoint_dir: Optional[str] = None,
        **kwargs,
    ):
        """Wandb logger fix to save checkpoints in wandb

        Accepts an additional checkpoint_dir argument, pointing to the real checkpoint directory

        Args:
            name (Optional[str]): Display name for the run. Defaults to None.
            save_dir (Optional[str]): Path where data is saved. Defaults to None.
            offline (Optional[bool]): Run offline (data can be streamed later to wandb servers). Defaults to False.
            id (Optional[str]): Sets the version, mainly used to resume a previous run. Defaults to None.
            anonymous (Optional[bool]): Enables or explicitly disables anonymous logging. Defaults to False.
            version (Optional[str]): Sets the version, mainly used to resume a previous run. Defaults to None.
            project (Optional[str]): The name of the project to which this run will belong. Defaults to None.
            log_model (Optional[bool]): Save checkpoints in wandb dir to upload on W&B servers. Defaults to False.
            experiment ([type]): WandB experiment object. Defaults to None.
            prefix (Optional[str]): A string to put at the beginning of metric keys. Defaults to "".
            sync_step (Optional[bool]): Sync Trainer step with wandb step. Defaults to True.
            checkpoint_dir (Optional[str]): Real checkpoint dir. Defaults to None.
        """
        self._checkpoint_dir = checkpoint_dir
        super(FixedWandbLogger, self).__init__(
            name=name,
            save_dir=save_dir,
            offline=offline,
            id=id,
            anonymous=anonymous,
            version=version,
            project=project,
            log_model=log_model,
            experiment=experiment,
            prefix=prefix,
            sync_step=sync_step,
            **kwargs,
        )

    @rank_zero_only
    def finalize(self, status: str) -> None:
        """Determine where checkpoints are saved and upload to wandb servers

        Args:
            status (str): Experiment status
        """
        # offset future training logged on same W&B run

        if self._experiment is not None:
            self._step_offset = self._experiment.step

        checkpoint_dir = (
            self._checkpoint_dir if self._checkpoint_dir is not None else self.save_dir
        )

        if checkpoint_dir is None:
            logger.warning(
                "Invalid checkpoint dir. Checkpoints will not be uploaded to Wandb."
            )
            logger.info(
                "You can manually upload your checkpoints through the CLI interface."
            )

        else:
            # upload all checkpoints from saving dir

            if self._log_model:
                wandb.save(os.path.join(checkpoint_dir, "*.ckpt"))



import math

import torch.nn as nn
from slp.modules.attention import MultiheadAttention
from slp.modules.embed import Embed, PositionalEncoding
from slp.modules.feedforward import PositionwiseFF
from slp.modules.norm import LayerNorm
from slp.util.pytorch import repeat_layer


class Sublayer1(nn.Module):
    def __init__(self, hidden_size=512, num_heads=8, dropout=0.1):
        super(Sublayer1, self).__init__()
        self.lnorm = LayerNorm(hidden_size)
        self.sublayer = MultiheadAttention(
            attention_size=hidden_size, num_heads=num_heads, dropout=dropout
        )

    def forward(self, x, attention_mask=None):
        return self.lnorm(x + self.sublayer(x, attention_mask=attention_mask))


class Sublayer2(nn.Module):
    def __init__(self, hidden_size=512, inner_size=2048, dropout=0.1):
        super(Sublayer2, self).__init__()
        self.lnorm = LayerNorm(hidden_size)
        self.sublayer = PositionwiseFF(hidden_size, inner_size, dropout=dropout)

    def forward(self, x):
        return self.lnorm(x + self.sublayer(x))


class Sublayer3(nn.Module):
    def __init__(self, hidden_size=512, num_heads=8, dropout=0.1):
        super(Sublayer3, self).__init__()
        self.lnorm = LayerNorm(hidden_size)
        self.sublayer = MultiheadAttention(
            attention_size=hidden_size, num_heads=num_heads, dropout=dropout
        )

    def forward(self, x, y, attention_mask=None):
        return self.lnorm(
            x + self.sublayer(x, queries=y, values=x, attention_mask=attention_mask)
        )


class EncoderLayer(nn.Module):
    def __init__(self, hidden_size=512, num_heads=8, inner_size=2048, dropout=0.1):
        super(EncoderLayer, self).__init__()
        self.l1 = Sublayer1(
            hidden_size=hidden_size, num_heads=num_heads, dropout=dropout
        )
        self.l2 = Sublayer2(
            hidden_size=hidden_size, inner_size=inner_size, dropout=dropout
        )

    def forward(self, x, attention_mask=None):
        out = self.l1(x, attention_mask=attention_mask)
        out = self.l2(out)
        return out


class Encoder(nn.Module):
    def __init__(
        self, num_layers=6, hidden_size=512, num_heads=8, inner_size=2048, dropout=0.1
    ):
        super(Encoder, self).__init__()
        self.encoder = nn.ModuleList(
            repeat_layer(
                EncoderLayer(
                    hidden_size=hidden_size,
                    num_heads=num_heads,
                    inner_size=inner_size,
                    dropout=dropout,
                ),
                num_layers,
            )
        )

    def forward(self, x, attention_mask=None):
        for layer in self.encoder:
            x = layer(x, attention_mask=attention_mask)
        return x


class DecoderLayer(nn.Module):
    def __init__(self, hidden_size=512, num_heads=8, inner_size=2048, dropout=0.1):
        super(DecoderLayer, self).__init__()
        self.in_layer = Sublayer1(
            hidden_size=hidden_size, num_heads=num_heads, dropout=dropout
        )
        self.fuse_layer = Sublayer3(
            hidden_size=hidden_size, num_heads=num_heads, dropout=dropout
        )
        self.out_layer = Sublayer2(
            hidden_size=hidden_size, inner_size=inner_size, dropout=dropout
        )

    def forward(self, targets, encoded, source_mask=None, target_mask=None):
        targets = self.in_layer(targets, attention_mask=target_mask)
        out = self.fuse_layer(encoded, targets, attention_mask=source_mask)
        out = self.out_layer(out)
        return out


class Decoder(nn.Module):
    def __init__(
        self, num_layers=6, hidden_size=512, num_heads=8, inner_size=2048, dropout=0.1
    ):
        super(Decoder, self).__init__()
        self.decoder = nn.ModuleList(
            repeat_layer(
                DecoderLayer(
                    hidden_size=hidden_size,
                    num_heads=num_heads,
                    inner_size=inner_size,
                    dropout=dropout,
                ),
                num_layers,
            )
        )

    def forward(self, target, encoded, source_mask=None, target_mask=None):

        for l in self.decoder:
            target = l(
                target, encoded, source_mask=source_mask, target_mask=target_mask
            )
        return target


class EncoderDecoder(nn.Module):
    def __init__(
        self, num_layers=6, hidden_size=512, num_heads=8, inner_size=2048, dropout=0.1
    ):
        super(EncoderDecoder, self).__init__()
        self.encoder = Encoder(
            num_layers=num_layers,
            hidden_size=hidden_size,
            num_heads=num_heads,
            inner_size=inner_size,
            dropout=dropout,
        )
        self.decoder = Decoder(
            num_layers=num_layers,
            hidden_size=hidden_size,
            num_heads=num_heads,
            inner_size=inner_size,
            dropout=dropout,
        )

    def forward(self, source, target, source_mask=None, target_mask=None):
        encoded = self.encoder(source, attention_mask=source_mask)
        decoded = self.decoder(
            target, encoded, source_mask=source_mask, target_mask=target_mask
        )
        return decoded


class Transformer(nn.Module):
    def __init__(
        self,
        vocab_size=30000,
        max_length=256,
        num_layers=6,
        hidden_size=512,
        num_heads=8,
        inner_size=2048,
        dropout=0.1,
    ):
        super(Transformer, self).__init__()
        self.embed = Embed(
            vocab_size,
            hidden_size,
            scale=math.sqrt(hidden_size),
            dropout=dropout,
            trainable=True,
        )
        self.pe = PositionalEncoding(embedding_dim=hidden_size, max_len=max_length)
        self.transformer_block = EncoderDecoder(
            num_layers=num_layers,
            hidden_size=hidden_size,
            num_heads=num_heads,
            inner_size=inner_size,
            dropout=dropout,
        )
        self.drop = nn.Dropout(dropout)
        self.predict = nn.Linear(hidden_size, vocab_size)
        self._reset_parameters()

    def forward(self, source, target, source_mask=None, target_mask=None):
        source = self.embed(source)
        target = self.embed(target)
        # Adding embeddings + pos embeddings
        # is done in PositionalEncoding class
        source = self.pe(source)
        target = self.pe(target)
        out = self.transformer_block(
            source, target, source_mask=source_mask, target_mask=target_mask
        )
        out = self.drop(out)
        out = self.predict(out)
        return out

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)



import math
from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


def attention_scores(
    k: torch.Tensor,
    q: torch.Tensor,
    dk: int,
    attention_mask: Optional[torch.Tensor] = None,
    dropout: float = 0.2,
    training: bool = True,
) -> torch.Tensor:
    """Calculate attention scores for scaled dot product attention

    $$s = softmax(\\frac{Q \cdot K^T}{\sqrt{d}})$$

    * B: Batch size
    * L: Keys Sequence length
    * M: Queries Sequence length
    * H: Number of heads
    * A: Feature dimension

    Args:
        k (torch.Tensor): Single head [B, L, A] or multi-head [B, H, L, A/H] Keys tensor
        q (torch.Tensor): Single head [B, M, A] or multi-head [B, H, M, A/H] Keys tensor
        dk (int): Model dimension
        attention_mask (Optional[torch.Tensor]): Optional [B, L] pad mask or [B, M, L] pad mask + subsequent mask
            tensor with zeros in sequence indices that should be masked and ones in sequence indices that should be
            preserved. Defaults to None.
        dropout (float): Drop probability. Defaults to 0.2.
        training (bool): Is module in training phase? Defaults to True.

    Returns:
        torch.Tensor: [B, M, L] or [B, H, M, L] attention scores
    """
    scores = torch.matmul(q, k.transpose(-1, -2)) / math.sqrt(dk)

    if attention_mask is not None:
        scores = scores + ((1 - attention_mask.unsqueeze(1)) * -1e5)
    scores = F.softmax(scores, dim=-1)
    scores = F.dropout(scores, p=dropout, training=training)

    return scores


class Attention(nn.Module):
    def __init__(
        self,
        attention_size: int = 512,
        input_size: Optional[int] = None,
        dropout: float = 0.1,
    ):
        """Single-Headed Dot-product attention module

        Args:
            attention_size (int): Number of hidden features. Defaults to 512.
            input_size (Optional[int]): Input features. Defaults to None.
                If None input_size is set to attention_size.
            dropout (float): Drop probability. Defaults to 0.1.
        """
        super(Attention, self).__init__()

        if input_size is None:
            input_size = attention_size
        self.dk = input_size
        self.k = nn.Linear(input_size, attention_size, bias=False)
        self.q = nn.Linear(input_size, attention_size, bias=False)
        self.v = nn.Linear(input_size, attention_size, bias=False)
        self.dropout = dropout
        self._reset_parameters()

    def forward(
        self,
        keys: torch.Tensor,
        queries: Optional[torch.Tensor] = None,
        values: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Single-head scaled dot-product attention forward pass

        Outputs the values, where features for each sequence element are weighted by their respective attention scores

        $$a = softmax(\\frac{Q}{K^T}){\sqrt{d}}) \dot V$$

        * B: Batch size
        * L: Keys Sequence length
        * M: Queries Sequence length
        * H: Number of heads
        * A: Feature dimension

        Args:
            keys (torch.Tensor): [B, L, D] Keys tensor
            queries (Optional[torch.Tensor]): Optional [B, M, D] Queries tensor. If None queries = keys. Defaults to None.
            values (Optional[torch.Tensor]): Optional [B, L, D] Values tensor. If None values = keys. Defaults to None.
            attention_mask (Optional[torch.Tensor]): Optional [B, M, L] zero-one mask for sequence elements. Defaults to None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (Reweighted values [B, L, D], attention scores [B, M, L])
        """

        if queries is None:
            queries = keys

        if values is None:
            values = keys
        k = self.k(keys)  # (B, L, A)
        q = self.q(queries)  # (B, L, A)
        v = self.v(values)  # (B, L, A)

        # weights => (B, L, L)
        scores = attention_scores(
            k,
            q,
            self.dk,
            attention_mask=attention_mask,
            dropout=self.dropout,
            training=self.training,
        )

        # out => (B, L, A)
        out = torch.bmm(scores, v)

        return out, scores

    def _reset_parameters(self):
        """xavier uniform init for Linear layer weights"""
        nn.init.xavier_uniform_(self.k.weight)
        nn.init.xavier_uniform_(self.q.weight)
        nn.init.xavier_uniform_(self.v.weight)


class MultiheadAttention(nn.Module):
    def __init__(
        self,
        attention_size: int = 512,
        num_heads: int = 8,
        input_size: Optional[int] = None,
        dropout: float = 0.1,
    ):
        """Multi-Headed Dot-product attention module

        Args:
            attention_size (int): Number of hidden features. Defaults to 512.
            num_heads (int): Number of attention heads
            input_size (Optional[int]): Input features. Defaults to None.
                If None input_size is set to attention_size.
            dropout (float): Drop probability. Defaults to 0.1.
        """
        super(MultiheadAttention, self).__init__()

        if input_size is None:
            input_size = attention_size
        self.dk = input_size
        self.num_heads = num_heads
        self.head_size = int(attention_size / num_heads)
        self.attention_size = attention_size
        self.k = nn.Linear(input_size, attention_size, bias=False)
        self.q = nn.Linear(input_size, attention_size, bias=False)
        self.v = nn.Linear(input_size, attention_size, bias=False)
        self.output = nn.Linear(attention_size, attention_size)
        self.dropout = dropout
        self._reset_parameters()

    def _split_heads(self, x: torch.Tensor) -> torch.Tensor:
        """Split input tensor into multiple attention heads

        (Batch size, Length, Attention size) => (Batch size, Heads, Lengths, Attention size / Heads)

        Args:
            x (torch.Tensor): [B, L, A] input tensor

        Returns:
            torch.Tensor: [B, H, L, A/H] Splitted / reshaped tensor
        """
        batch_size, max_length, _ = x.size()

        return x.view(batch_size, max_length, self.num_heads, self.head_size).permute(
            0, 2, 1, 3
        )

    def _merge_heads(self, x: torch.Tensor) -> torch.Tensor:
        """Merge multiple attention heads into output tensor

        (Batch size, Heads, Lengths, Attention size / Heads) => (Batch size, Length, Attention size)

        Args:
            x (torch.Tensor): [B, H, L, A/H] multi-head tensor

        Returns:
            torch.Tensor:  [B, L, A] merged / reshaped tensor
        """
        batch_size, _, max_length, _ = x.size()
        # x => (B, L, H, A/H)
        x = x.permute(0, 2, 1, 3).contiguous()

        return x.view(batch_size, max_length, -1)

    def forward(self, keys, queries=None, values=None, attention_mask=None):
        """Multi-head scaled dot-product attention forward pass

        Outputs the values, where features for each sequence element are weighted by their respective attention scores

        Each head performs dot-product attention

        $$a_H = softmax(\\frac{Q_H \cdot K_H^T}{\sqrt{d}}) \cdot V_H$$

        The outputs of multiple heads are concatenated and passed through a feedforward layer.

        $$a = W (a^{(1)}_{H} \mathbin\Vert a^{(2)}_{H} \dots) + b$$


        * B: Batch size
        * L: Keys Sequence length
        * M: Queries Sequence length
        * H: Number of heads
        * A: Feature dimension


        Args:
            keys (torch.Tensor): [B, L, D] Keys tensor
            queries (Optional[torch.Tensor]): Optional [B, M, D] Queries tensor. If None queries = keys. Defaults to None.
            values (Optional[torch.Tensor]): Optional [B, L, D] Values tensor. If None values = keys. Defaults to None.
            attention_mask (Optional[torch.Tensor]): Optional [B, M, L] zero-one mask for sequence elements. Defaults to None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (Reweighted values [B, L, D], attention scores [B, H, M, L])
        """
        if queries is None:
            queries = keys

        if values is None:
            values = keys
        k = self._split_heads(self.k(keys))  # (B, H, L, A/H)
        q = self._split_heads(self.q(queries))  # (B, H, L, A/H)
        v = self._split_heads(self.v(values))  # (B, H, L, A/H)

        # scores => (B, H, L, L)
        scores = attention_scores(
            k,
            q,
            self.dk,
            attention_mask=attention_mask,
            dropout=self.dropout,
            training=self.training,
        )

        # out => (B, H, L, A/H)
        out = self._merge_heads(torch.matmul(scores, v))
        out = self.output(out)

        return out

    def _reset_parameters(self):
        """xavier uniform init for Linear layer weights"""
        nn.init.xavier_uniform_(self.k.weight)
        nn.init.xavier_uniform_(self.q.weight)
        nn.init.xavier_uniform_(self.v.weight)
        nn.init.xavier_uniform_(self.output.weight)
        nn.init.constant_(self.output.bias, 0.0)



import itertools
from typing import List, Union

import torch
import torch.nn as nn
from loguru import logger
from slp.modules.norm import LayerNorm

# Activation functions to choose
NON_LINEARITIES = {
    "relu": nn.ReLU,
    "tanh": nn.Tanh,
    "sigmoid": nn.Sigmoid,
    "none": None,
}


class FF(nn.Module):
    def __init__(
        self,
        n_in: int,
        n_out: int,
        activation: str = "relu",
        layer_norm: bool = True,
        bias: bool = True,
        dropout: float = 0.1,
    ):
        """Feed-forward layer with optional layernorm, and activation

        Args:
            n_in (int): Input features
            n_out (int): Output features
            activation (str): Which activation to use. Defaults to "relu".
            layer_norm (bool): Use layernorm. Defaults to True.
            bias (bool): Add bias. Defaults to True.
            dropout (float): Dropout probability. Defaults to 0.1.
        """
        super(FF, self).__init__()
        self.fc = nn.Linear(n_in, n_out, bias=bias)
        self.activation = NON_LINEARITIES.get(activation, nn.ReLU)
        if self.activation is not None:
            self.activation = self.activation()
        self.layer_norm = None
        if layer_norm:
            self.layer_norm = LayerNorm(n_out)
        self.drop = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward FF pass

        Args:
            x (torch.Tensor): [B, *, D] input features

        Returns:
            torch.Tensor: [B, *, D] output features
        """
        out = self.fc(x)
        out = self.drop(out)
        if self.layer_norm is not None:
            out = self.layer_norm(out)
        if self.activation is not None:
            out = self.activation(out)
        return out  # type: ignore


class PositionwiseFF(nn.Module):
    """Some Information about PositionwiseFF"""

    def __init__(self, d_model: int, d_ff: int, dropout: float = 0.1):
        """Transformer Position-wise feed-forward layer

        Linear -> LayerNorm -> ReLU -> Linear

        Args:
            d_model (int): Model dimension
            d_ff (int): Hidden dimension
            dropout (float): Dropout probability. Defaults to 0.1.
        """
        super(PositionwiseFF, self).__init__()
        self.ff1 = FF(d_model, d_ff, activation="relu")
        self.ff2 = nn.Linear(d_ff, d_model)
        self.drop = nn.Dropout(dropout)
        self.net = nn.Sequential(self.ff1, self.drop, self.ff2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Position-wise FF forward pass

        $$out = W_2 \dot max(0, W_1 \dot x + b_1) + b_2$$

        [B, *, D] -> [B, *, H] -> [B, *, D]

        * B: Batch size
        * D: Model dim
        * H: Hidden size > Model dim (Usually $H = 2D$)

        Args:
            x (torch.Tensor): [B, *, D] Input features

        Returns:
            torch.Tensor: [B, *, D] Output features
        """
        out: torch.Tensor = self.net(x)
        return out


class MultilayerFF(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        sizes: Union[int, List[int]],
        n_layers: int = 1,
        activation: str = "relu",
        dropout: float = 0.1,
    ):
        """Multi-layer feed forward layer.

        Stacks multiple FFs with configurable dimensions

        Args:
            input_dim (int): Input features
            output_dim (int): Output features
            sizes (Union[int, List[int]]): Intermediate layer features as a list of dimensions.
                If single int is passed all intermediate layers have the same dimension
            n_layers (int): Number of intermediate layers. Defaults to 1.
            activation (str): Activation function. Defaults to "relu".
            dropout (float): Dropout probability. Defaults to 0.1.
        """
        super(MultilayerFF, self).__init__()
        if isinstance(sizes, int):
            sizes = list(itertools.repeat(sizes, n_layers))  # [n] * l
        sizes = [input_dim] + sizes + [output_dim]
        if len(sizes) != n_layers + 2:
            logger.warning(
                f"n_layers={n_layers} does not match len of "
                "sizes={len(sizes)}. Using {len(sizes)} layers"
            )
        self.net = nn.Sequential(
            *[
                FF(nin, nout, activation=activation, dropout=dropout)
                for nin, nout in zip(sizes[:-1], sizes[1:])
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward Multilayer-FF pass

        Args:
            x (torch.Tensor): [B, *, D] input features

        Returns:
            torch.Tensor: [B, *, D] output features
        """
        out: torch.Tensor = self.net(x)
        return out



from typing import List, Tuple

import torch
from slp.util.pytorch import mktensor
from slp.util.types import Label
from torch.nn.utils.rnn import pad_sequence


class SequenceClassificationCollator(object):
    def __init__(self, pad_indx=0, device="cpu"):
        """Collate function for sequence classification tasks

        * Perform padding
        * Calculate sequence lengths

        Args:
            pad_indx (int): Pad token index. Defaults to 0.
            device (str): device of returned tensors. Leave this as "cpu".
                The LightningModule will handle the Conversion.

        Examples:
        >>> dataloader = torch.utils.DataLoader(my_dataset, collate_fn=SequenceClassificationCollator())
        """
        self.pad_indx = pad_indx
        self.device = device

    def __call__(
        self, batch: List[Tuple[torch.Tensor, Label]]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Call collate function

        Args:
            batch (List[Tuple[torch.Tensor, slp.util.types.Label]]): Batch of samples.
                It expects a list of tuples (inputs, label).

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: Returns tuple of batched tensors (inputs, labels, lengths)
        """
        inputs: List[torch.Tensor] = [b[0] for b in batch]
        targets: List[Label] = [b[1] for b in batch]
        #  targets: List[torch.tensor] = map(list, zip(*batch))
        lengths = torch.tensor([s.size(0) for s in inputs], device=self.device)
        # Pad and convert to tensor
        inputs_padded: torch.Tensor = pad_sequence(
            inputs, batch_first=True, padding_value=self.pad_indx
        ).to(self.device)
        ttargets: torch.Tensor = mktensor(targets, device=self.device, dtype=torch.long)
        return inputs_padded, ttargets.to(self.device), lengths


class Seq2SeqCollator(object):
    def __init__(self, pad_indx=0, device="cpu"):
        """Collate function for seq2seq tasks

        * Perform padding
        * Calculate sequence lengths

        Args:
            pad_indx (int): Pad token index. Defaults to 0.
            device (str): device of returned tensors. Leave this as "cpu".
                The LightningModule will handle the Conversion.

        Examples:
        >>> dataloader = torch.utils.DataLoader(my_dataset, collate_fn=Seq2SeqClassificationCollator())
        """
        self.pad_indx = pad_indx
        self.device = device

    def __call__(
        self, batch: List[Tuple[torch.Tensor, torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Call collate function

        Args:
            batch (List[Tuple[torch.Tensor, torch.Tensor]]): Batch of samples.
                It expects a list of tuples (source, target)
                Each source and target are a sequences of features or ids.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: Returns tuple of batched tensors
                (inputs, labels, lengths_inputs, lengths_targets)
        """
        inputs: List[torch.Tensor] = [b[0] for b in batch]
        targets: List[torch.Tensor] = [b[1] for b in batch]
        lengths_inputs = torch.tensor([s.size(0) for s in inputs], device=self.device)
        lengths_targets = torch.tensor([s.size(0) for s in targets], device=self.device)

        inputs_padded: torch.Tensor = pad_sequence(
            inputs, batch_first=True, padding_value=self.pad_indx
        ).to(self.device)

        targets_padded: torch.Tensor = pad_sequence(
            targets, batch_first=True, padding_value=self.pad_indx
        ).to(self.device)

        return inputs_padded, targets_padded, lengths_inputs, lengths_targets
